import numpy
import pytest

from nabu.stitching.overlap import compute_image_minimum_divergence, compute_image_higher_signal, check_overlaps
from nabu.testutils import get_data
from nabu.stitching.overlap import ImageStichOverlapKernel, OverlapStitchingStrategy
from nabu.stitching.stitcher_2D import stitch_raw_frames
from silx.image.phantomgenerator import PhantomGenerator


strategies_to_test_weights = (
    OverlapStitchingStrategy.CLOSEST,
    OverlapStitchingStrategy.COSINUS_WEIGHTS,
    OverlapStitchingStrategy.LINEAR_WEIGHTS,
    OverlapStitchingStrategy.MEAN,
)


@pytest.mark.parametrize("strategy", strategies_to_test_weights)
@pytest.mark.parametrize("stitching_axis", (0, 1))
def test_overlap_stitcher(strategy, stitching_axis):
    frame_width = 128
    frame_height = frame_width
    frame_1 = PhantomGenerator.get2DPhantomSheppLogan(n=frame_width)
    stitcher = ImageStichOverlapKernel(
        stitching_strategy=strategy,
        overlap_size=frame_height,
        frame_unstitched_axis_size=128,
        stitching_axis=stitching_axis,
    )
    stitched_frame = stitcher.stitch(frame_1, frame_1)[0]
    assert stitched_frame.shape == (frame_height, frame_width)
    # check result is close to the expected one
    numpy.testing.assert_allclose(frame_1, stitched_frame, atol=10e-10)

    # check sum of weights ~ 1.0
    numpy.testing.assert_allclose(
        stitcher.weights_img_1 + stitcher.weights_img_2,
        numpy.ones_like(stitcher.weights_img_1),
    )


@pytest.mark.parametrize("stitching_axis", (0, 1))
def test_compute_image_minimum_divergence(stitching_axis):
    """make sure the compute_image_minimum_divergence function is processing"""
    raw_data_1 = get_data("brain_phantom.npz")["data"]
    raw_data_2 = numpy.random.rand(*raw_data_1.shape) * 255.0

    stitching = compute_image_minimum_divergence(
        raw_data_1, raw_data_2, high_frequency_threshold=2, stitching_axis=stitching_axis
    )
    assert stitching.shape == raw_data_1.shape


def test_compute_image_higher_signal():
    """
    make sure compute_image_higher_signal is processing
    """
    raw_data = get_data("brain_phantom.npz")["data"]
    raw_data_1 = raw_data.copy()
    raw_data_1[40:75] = 0.0
    raw_data_1[:, 210:245] = 0.0

    raw_data_2 = raw_data.copy()
    raw_data_2[:, 100:120] = 0.0

    stitching = compute_image_higher_signal(raw_data_1, raw_data_2)

    numpy.testing.assert_array_equal(
        stitching,
        raw_data,
    )


def test_check_overlaps():
    """test 'check_overlaps' function"""

    # two frames, ordered and with an overlap
    check_overlaps(
        frames=(
            numpy.ones(10),
            numpy.ones(20),
        ),
        positions=((10, 0, 0), (0, 0, 0)),
        axis=0,
        raise_error=True,
    )

    # two frames, ordered and without an overlap
    with pytest.raises(ValueError):
        check_overlaps(
            frames=(
                numpy.ones(10),
                numpy.ones(20),
            ),
            positions=((0, 0, 0), (100, 0, 0)),
            axis=0,
            raise_error=True,
        )

    # two frames, frame 0 fully overlap frame 1
    with pytest.raises(ValueError):
        check_overlaps(
            frames=(
                numpy.ones(20),
                numpy.ones(10),
            ),
            positions=((8, 0, 0), (5, 0, 0)),
            axis=0,
            raise_error=True,
        )

    # three frames 'overlaping' as expected
    check_overlaps(
        frames=(
            numpy.ones(10),
            numpy.ones(20),
            numpy.ones(10),
        ),
        positions=((20, 0, 0), (10, 0, 0), (0, 0, 0)),
        axis=0,
        raise_error=True,
    )

    # three frames: frame 0 overlap frame 1 but also frame 2
    with pytest.raises(ValueError):
        check_overlaps(
            frames=(
                numpy.ones(20),
                numpy.ones(10),
                numpy.ones(10),
            ),
            positions=((20, 0, 0), (15, 0, 0), (11, 0, 0)),
            axis=0,
            raise_error=True,
        )


@pytest.mark.parametrize("dtype", (numpy.float16, numpy.float32))
def test_stitch_vertically_raw_frames(dtype):
    """
    ensure a stitching with 3 frames and different overlap can be done
    """
    ref_frame_width = 256
    frame_ref = PhantomGenerator.get2DPhantomSheppLogan(n=ref_frame_width).astype(dtype)

    # split the frame into several part
    frame_1 = frame_ref[0:100]
    frame_2 = frame_ref[80:164]
    frame_3 = frame_ref[154:]

    kernel_1 = ImageStichOverlapKernel(frame_unstitched_axis_size=ref_frame_width, overlap_size=20, stitching_axis=0)
    kernel_2 = ImageStichOverlapKernel(frame_unstitched_axis_size=ref_frame_width, overlap_size=10, stitching_axis=0)

    stitched = stitch_raw_frames(
        frames=(frame_1, frame_2, frame_3),
        output_dtype=dtype,
        overlap_kernels=(kernel_1, kernel_2),
        raw_frames_compositions=None,
        overlap_frames_compositions=None,
        key_lines=(
            (
                90,  # frame_1 height - kernel_1 height / 2.0
                10,  # kernel_1 height / 2.0
            ),
            (
                79,  # frame_2 height - kernel_2 height / 2.0 ou 102-20 ?
                5,  # kernel_2 height / 2.0
            ),
        ),
    )

    assert stitched.shape == frame_ref.shape
    numpy.testing.assert_array_almost_equal(frame_ref, stitched)


def test_stitch_vertically_raw_frames_2():
    """
    ensure a stitching with 3 frames and different overlap can be done
    """
    ref_frame_width = 256
    frame_ref = PhantomGenerator.get2DPhantomSheppLogan(n=ref_frame_width).astype(numpy.float32)

    # split the frame into several part
    frame_1 = frame_ref.copy()
    frame_2 = frame_ref.copy()
    frame_3 = frame_ref.copy()

    kernel_1 = ImageStichOverlapKernel(frame_unstitched_axis_size=ref_frame_width, overlap_size=10, stitching_axis=0)
    kernel_2 = ImageStichOverlapKernel(frame_unstitched_axis_size=ref_frame_width, overlap_size=10, stitching_axis=0)

    stitched = stitch_raw_frames(
        frames=(frame_1, frame_2, frame_3),
        output_dtype=numpy.float32,
        overlap_kernels=(kernel_1, kernel_2),
        raw_frames_compositions=None,
        overlap_frames_compositions=None,
        key_lines=((20, 20), (105, 105)),
    )

    assert stitched.shape == frame_ref.shape
    numpy.testing.assert_array_almost_equal(frame_ref, stitched)


@pytest.mark.parametrize("dtype", (numpy.float16, numpy.float32))
def test_stitch_horizontally_raw_frames(dtype):
    """
    ensure a stitching with 3 frames and different overlap can be done along axis 1
    """
    ref_frame_width = 256
    frame_ref = PhantomGenerator.get2DPhantomSheppLogan(n=ref_frame_width).astype(dtype)

    # split the frame into several part
    frame_1 = frame_ref[:, 0:100]
    frame_2 = frame_ref[:, 80:164]
    frame_3 = frame_ref[:, 154:]

    kernel_1 = ImageStichOverlapKernel(frame_unstitched_axis_size=ref_frame_width, overlap_size=20, stitching_axis=1)
    kernel_2 = ImageStichOverlapKernel(frame_unstitched_axis_size=ref_frame_width, overlap_size=10, stitching_axis=1)

    stitched = stitch_raw_frames(
        frames=(frame_1, frame_2, frame_3),
        output_dtype=dtype,
        overlap_kernels=(kernel_1, kernel_2),
        raw_frames_compositions=None,
        overlap_frames_compositions=None,
        key_lines=(
            (
                90,  # frame_1 height - kernel_1 height / 2.0
                10,  # kernel_1 height / 2.0
            ),
            (
                79,  # frame_2 height - kernel_2 height / 2.0 ou 102-20 ?
                5,  # kernel_2 height / 2.0
            ),
        ),
    )

    assert stitched.shape == frame_ref.shape
    numpy.testing.assert_array_almost_equal(frame_ref, stitched)



from ..processing.histogram_cuda import *
from ..utils import deprecation_warning

deprecation_warning(
    "nabu.misc.histogram_cuda has been moved to nabu.processing.histogram_cuda",
    do_print=True,
    func_name="histogram_cuda",
)



from dataclasses import dataclass
import numpy
from math import ceil


@dataclass
class FrameComposition:
    """
    class used to define intervals to know where to dump raw data or stitched data according to requested policy.
    The idea is to create this once for all for one stitching operation and reuse it for each frame.
    """

    composed_axis: int
    """axis along which the composition is done"""
    local_start: tuple
    """tuple of indices on the input frames ref to know where each region start (along the composed axis)"""
    local_end: tuple
    """tuple of indices on the input frames ref to know where each region end (along the composed axis)"""
    global_start: tuple
    """tuple of indices on the output frame ref to know where each region start (along the composed axis)"""
    global_end: tuple
    """tuple of indices on the output frame ref to know where each region end (along the composed axis)"""

    def browse(self):
        for i in range(len(self.local_start)):
            yield (
                self.local_start[i],
                self.local_end[i],
                self.global_start[i],
                self.global_end[i],
            )

    def compose(self, output_frame: numpy.ndarray, input_frames: tuple):
        if not output_frame.ndim in (2, 3):
            raise TypeError(
                f"output_frame is expected to be 2D (gray scale) or 3D (RGB(A)) and not {output_frame.ndim}"
            )
        for (
            global_start,
            global_end,
            local_start,
            local_end,
            input_frame,
        ) in zip(
            self.global_start,
            self.global_end,
            self.local_start,
            self.local_end,
            input_frames,
        ):
            if input_frame is not None:
                if self.composed_axis == 0:
                    output_frame[global_start:global_end] = input_frame[local_start:local_end]
                elif self.composed_axis == 1:
                    output_frame[:, global_start:global_end] = input_frame[:, local_start:local_end]
                else:
                    raise ValueError(f"composed axis must be in (0, 1). Get {self.composed_axis}")

    @staticmethod
    def compute_raw_frame_compositions(frames: tuple, key_lines: tuple, overlap_kernels: tuple, stitching_axis):
        """
        compute frame composition for raw data

        warning: we expect frames to be ordered y downward and the frame order to keep this ordering
        """
        assert len(frames) == len(overlap_kernels) + 1 == len(key_lines) + 1

        global_start_indices = [0]

        # extend shifts and kernels to have a first shift of 0 and two overlaps values at 0 to
        # generalize processing
        local_start_indices = [0]

        local_start_indices.extend(
            [ceil(key_line[1] + kernel.overlap_size / 2) for (key_line, kernel) in zip(key_lines, overlap_kernels)]
        )
        local_end_indices = list(
            [ceil(key_line[0] - kernel.overlap_size / 2) for (key_line, kernel) in zip(key_lines, overlap_kernels)]
        )
        local_end_indices.append(frames[-1].shape[stitching_axis])

        for (
            new_local_start_index,
            new_local_end_index,
            kernel,
        ) in zip(local_start_indices, local_end_indices, overlap_kernels):
            global_start_indices.append(
                global_start_indices[-1] + (new_local_end_index - new_local_start_index) + kernel.overlap_size
            )

        # global end can be easily found from global start + local start and end
        global_end_indices = []
        for global_start_index, new_local_start_index, new_local_end_index in zip(
            global_start_indices, local_start_indices, local_end_indices
        ):
            global_end_indices.append(global_start_index + new_local_end_index - new_local_start_index)

        return FrameComposition(
            composed_axis=stitching_axis,
            local_start=tuple(local_start_indices),
            local_end=tuple(local_end_indices),
            global_start=tuple(global_start_indices),
            global_end=tuple(global_end_indices),
        )

    @staticmethod
    def compute_stitch_frame_composition(frames, key_lines: tuple, overlap_kernels: tuple, stitching_axis: int):
        """
        compute frame composition for stiching.
        """
        assert len(frames) == len(overlap_kernels) + 1 == len(key_lines) + 1
        assert stitching_axis in (0, 1)

        # position in the stitched frame;
        local_start_indices = [0] * len(overlap_kernels)
        local_end_indices = [kernel.overlap_size for kernel in overlap_kernels]

        # position in the global frame. For this one it is simpler to rely on the raw frame composition
        composition_raw = FrameComposition.compute_raw_frame_compositions(
            frames=frames,
            key_lines=key_lines,
            overlap_kernels=overlap_kernels,
            stitching_axis=stitching_axis,
        )
        global_start_indices = composition_raw.global_end[:-1]
        global_end_indices = composition_raw.global_start[1:]

        return FrameComposition(
            composed_axis=stitching_axis,
            local_start=tuple(local_start_indices),
            local_end=tuple(local_end_indices),
            global_start=tuple(global_start_indices),
            global_end=tuple(global_end_indices),
        )

    @staticmethod
    def pprint_composition(raw_composition, stitch_composition):
        """
        util to display what the output of the composition will looks like from composition
        """
        for i_frame, (raw_comp, stitch_comp) in enumerate(zip(raw_composition.browse(), stitch_composition.browse())):
            raw_local_start, raw_local_end, raw_global_start, raw_global_end = raw_comp

            print(
                f"stitch_frame[{raw_global_start}:{raw_global_end}] = frame_{i_frame}[{raw_local_start}:{raw_local_end}]"
            )

            (
                stitch_local_start,
                stitch_local_end,
                stitch_global_start,
                stitch_global_end,
            ) = stitch_comp

            print(
                f"stitch_frame[{stitch_global_start}:{stitch_global_end}] = stitched_frame_{i_frame}[{stitch_local_start}:{stitch_local_end}]"
            )
        else:
            i_frame += 1
            raw_local_start, raw_local_end, raw_global_start, raw_global_end = list(raw_composition.browse())[-1]
            print(
                f"stitch_frame[{raw_global_start}:{raw_global_end}] = frame_{i_frame}[{raw_local_start}:{raw_local_end}]"
            )



import numpy as np
import pytest
from nabu.processing.fft_cuda import get_available_fft_implems
from nabu.testutils import (
    get_big_data,
    __big_testdata_dir__,
    generate_tests_scenarios,
    __do_long_tests__,
)
from nabu.cuda.utils import __has_pycuda__, get_cuda_context

__has_cufft__ = False
if __has_pycuda__:
    avail_fft = get_available_fft_implems()
    __has_cufft__ = len(avail_fft) > 0

__has_cuda_fbp__ = __has_cufft__ and __has_pycuda__
if __has_cuda_fbp__:
    from nabu.reconstruction.reconstructor_cuda import CudaReconstructor
    from nabu.reconstruction.fbp import Backprojector as CudaBackprojector
    import pycuda.gpuarray as garray


scenarios = generate_tests_scenarios(
    {
        "axis": ["z", "y", "x"],
        "vol_type": ["sinograms", "projections"],
        "indices": [(300, 310)],  # reconstruct 10 slices
        "slices_roi": [None, (None, None, 200, 400), (250, 300, None, None), (120, 330, 301, 512)],
    }
)


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    cls.projs = get_big_data("MRI512_projs.npy")
    if __has_cuda_fbp__:
        cls.ctx = get_cuda_context()
        cls.d_projs = garray.to_gpu(cls.projs)
    cls.ref = None
    cls.tol = 5e-2


@pytest.mark.skipif(
    __big_testdata_dir__ is None or not (__do_long_tests__),
    reason="need environment variable NABU_BIGDATA_DIR and NABU_LONG_TESTS=1",
)
@pytest.mark.usefixtures("bootstrap")
class TestReconstructor:
    @pytest.mark.skipif(not (__has_cuda_fbp__), reason="need pycuda and (scikit-cuda or vkfft)")
    @pytest.mark.parametrize("config", scenarios)
    def test_cuda_reconstructor(self, config):
        data = self.projs
        d_data = self.d_projs
        if config["vol_type"] == "sinograms":
            data = np.moveaxis(self.projs, 1, 0)
            d_data = self.d_projs.transpose(axes=(1, 0, 2))  # view
        reconstructor = CudaReconstructor(
            data.shape,
            config["indices"],
            axis=config["axis"],
            vol_type=config["vol_type"],
            slices_roi=config["slices_roi"],
        )
        res = reconstructor.reconstruct(d_data)
        ref = self.get_ref()
        ref = self.crop_array(ref, config)
        err_max = np.max(np.abs(res - ref))
        assert err_max < self.tol, "something wrong with reconstructor, config = %s" % str(config)

    def get_ref(self):
        if self.ref is not None:
            return self.ref
        if __has_cuda_fbp__:
            fbp_cls = CudaBackprojector
        ref = np.zeros((512, 512, 512), "f")
        fbp = fbp_cls((self.projs.shape[0], self.projs.shape[-1]))
        for i in range(512):
            ref[i] = fbp.fbp(self.d_projs[:, i, :])
        self.ref = ref
        return self.ref

    @staticmethod
    def crop_array(arr, config):
        indices = config["indices"]
        axis = config["axis"]
        slices_roi = config["slices_roi"] or (None, None, None, None)
        i_slice = slice(*indices)
        u_slice = slice(*slices_roi[:2])
        v_slice = slice(*slices_roi[-2:])
        if axis == "z":
            z_slice, y_slice, x_slice = i_slice, v_slice, u_slice
        if axis == "y":
            z_slice, y_slice, x_slice = v_slice, i_slice, u_slice
        if axis == "x":
            z_slice, y_slice, x_slice = v_slice, u_slice, i_slice
        return arr[z_slice, y_slice, x_slice]



import os
import pytest
import numpy as np
from nabu.testutils import utilstest, __do_long_tests__
from nabu.resources.dataset_analyzer import HDF5DatasetAnalyzer, analyze_dataset
from nabu.resources.nxflatfield import update_dataset_info_flats_darks
from nabu.resources.utils import extract_parameters
from nabu.pipeline.estimators import CompositeCOREstimator
from nabu.pipeline.config import parse_nabu_config_file
from nabu.pipeline.estimators import SinoCORFinder, CORFinder


#
# Test CoR estimation with "composite-coarse-to-fine" (aka "near" in the legacy system vocable)
#


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls

    dataset_downloaded_path = utilstest.getfile("test_composite_cor_finder_data.h5")
    cls.theta_interval = 4.5 * 1  # this is given. Radios in the middle of steps 4.5 degree long
    # are set to zero for compression
    # You can still change it to a multiple of 4.5
    cls.cor_pix = 1321.625
    cls.abs_tol = 0.0001
    cls.dataset_info = HDF5DatasetAnalyzer(dataset_downloaded_path)
    update_dataset_info_flats_darks(cls.dataset_info, True)
    cls.cor_options = extract_parameters("side=300.0;  near_width = 20.0", sep=";")


@pytest.mark.skipif(not (__do_long_tests__), reason="Need NABU_LONG_TESTS=1 for this test")
@pytest.mark.usefixtures("bootstrap")
class TestCompositeCorFinder:
    def test(self):
        cor_finder = CompositeCOREstimator(
            self.dataset_info, theta_interval=self.theta_interval, cor_options=self.cor_options
        )

        cor_position = cor_finder.find_cor()
        message = "Computed CoR %f " % cor_position + " and real CoR %f do not coincide" % self.cor_pix
        assert np.isclose(self.cor_pix, cor_position, atol=self.abs_tol), message


@pytest.fixture(scope="class")
def bootstrap_bamboo_reduced(request):
    cls = request.cls
    cls.abs_tol = 0.2
    # Dataset without estimated_cor_frm_motor (non regression test)
    dataset_relpath = os.path.join("bamboo_reduced.nx")
    dataset_downloaded_path = utilstest.getfile(dataset_relpath)
    conf_relpath = os.path.join("bamboo_reduced.conf")
    conf_downloaded_path = utilstest.getfile(conf_relpath)
    cls.ds_std = analyze_dataset(dataset_downloaded_path)
    update_dataset_info_flats_darks(cls.ds_std, True)
    cls.conf_std = parse_nabu_config_file(conf_downloaded_path)

    # Dataset with estimated_cor_frm_motor
    dataset_relpath = os.path.join("bamboo_reduced_bliss.nx")
    dataset_downloaded_path = utilstest.getfile(dataset_relpath)
    conf_relpath = os.path.join("bamboo_reduced_bliss.conf")
    conf_downloaded_path = utilstest.getfile(conf_relpath)
    cls.ds_bliss = analyze_dataset(dataset_downloaded_path)
    update_dataset_info_flats_darks(cls.ds_bliss, True)
    cls.conf_bliss = parse_nabu_config_file(conf_downloaded_path)


@pytest.mark.skipif(not (__do_long_tests__), reason="need environment variable NABU_LONG_TESTS=1")
@pytest.mark.usefixtures("bootstrap_bamboo_reduced")
class TestCorNearPos:
    # TODO adapt test file
    true_cor = 339.486 - 0.5

    def test_cor_sliding_standard(self):
        cor_options = extract_parameters(self.conf_std["reconstruction"].get("cor_options", None), sep=";")
        for side in [None, "from_file", "center"]:
            if side is not None:
                cor_options.update({"side": side})
            finder = CORFinder("sliding-window", self.ds_std, do_flatfield=True, cor_options=cor_options)
            cor = finder.find_cor()
            message = f"Computed CoR {cor} and expected CoR {self.true_cor} do not coincide. Near_pos options was set to {cor_options.get('near_pos',None)}."
            assert np.isclose(self.true_cor, cor, atol=self.abs_tol + 0.5), message  # FIXME

    def test_cor_fourier_angles_standard(self):
        cor_options = extract_parameters(self.conf_std["reconstruction"].get("cor_options", None), sep=";")
        # TODO modify test files
        if "near_pos" in cor_options and "near" in cor_options.get("side", "") == "near":
            cor_options["side"] = cor_options["near_pos"]
        #
        for side in [None, "from_file", "center"]:
            if side is not None:
                cor_options.update({"side": side})
            finder = SinoCORFinder("fourier-angles", self.ds_std, do_flatfield=True, cor_options=cor_options)
            cor = finder.find_cor()
            message = f"Computed CoR {cor} and expected CoR {self.true_cor} do not coincide. Near_pos options was set to {cor_options.get('near_pos',None)}."
            assert np.isclose(self.true_cor + 0.5, cor, atol=self.abs_tol), message

    def test_cor_sliding_bliss(self):
        cor_options = extract_parameters(self.conf_bliss["reconstruction"].get("cor_options", None), sep=";")
        # TODO modify test files
        if "near_pos" in cor_options and "near" in cor_options.get("side", "") == "near":
            cor_options["side"] = cor_options["near_pos"]
        #
        for side in [None, "from_file", "center"]:
            if side is not None:
                cor_options.update({"side": side})
            finder = CORFinder("sliding-window", self.ds_bliss, do_flatfield=True, cor_options=cor_options)
            cor = finder.find_cor()
            message = f"Computed CoR {cor} and expected CoR {self.true_cor} do not coincide. Near_pos options was set to {cor_options.get('near_pos',None)}."
            assert np.isclose(self.true_cor, cor, atol=self.abs_tol), message

    def test_cor_fourier_angles_bliss(self):
        cor_options = extract_parameters(self.conf_bliss["reconstruction"].get("cor_options", None), sep=";")
        for side in [None, "from_file", "center"]:
            if side is not None:
                cor_options.update({"side": side})
            finder = SinoCORFinder("fourier-angles", self.ds_bliss, do_flatfield=True, cor_options=cor_options)
            cor = finder.find_cor()
            message = f"Computed CoR {cor} and expected CoR {self.true_cor} do not coincide. Near_pos options was set to {cor_options.get('near_pos',None)}."
            assert np.isclose(self.true_cor + 0.5, cor, atol=self.abs_tol), message



import os

path = os.path
from ..utils import check_supported, is_writeable
from .params import *

"""
A validator is a function with
  - input: a value
  - output: the input value, or a modified input value
  - possibly raising exceptions in case of invalid value.
"""


# ------------------------------------------------------------------------------
# ---------------------------- Utils -------------------------------------------
# ------------------------------------------------------------------------------


def raise_error(section, key, msg=""):
    raise ValueError("Invalid value for %s/%s: %s" % (section, key, msg))


def validator(func):
    """
    Common decorator for all validator functions.
    It modifies the signature of the decorated functions !
    """

    def wrapper(section, key, value):
        try:
            res = func(value)
        except AssertionError as e:
            raise_error(section, key, e)
        return res

    return wrapper


def convert_to_int(val):
    val_int = 0
    try:
        val_int = int(val)
        conversion_error = None
    except ValueError as exc:
        conversion_error = exc
    return val_int, conversion_error


def convert_to_float(val):
    val_float = 0.0
    try:
        val_float = float(val)
        conversion_error = None
    except ValueError as exc:
        conversion_error = exc
    return val_float, conversion_error


def convert_to_bool(val):
    val_int, error = convert_to_int(val)
    res = None
    if not error:
        res = val_int > 0
    else:
        if val.lower() in ["yes", "true", "y"]:
            res = True
            error = None
        if val.lower() in ["no", "false", "n"]:
            res = False
            error = None
    return res, error


def str2bool(val):
    """This is an interface to convert_to_bool and it is meant
    to work as a class: in argparse interface the type argument can be set to float, int .. in general to a class.
    The argument value is then created, at parsing time, by typecasting the input string to the given class.
    A possibly occuring exception then trigger, in case, the display explanation provided by the argparse library.
    All what this methods does is simply trying to convert an argument into a bool, and return it,
    or generate an exception if there is a problem
    """
    import argparse

    res, error = convert_to_bool(val)
    if error:
        raise argparse.ArgumentTypeError(error)
    else:
        return res


def convert_to_bool_noerr(val):
    res, err = convert_to_bool(val)
    if err is not None:
        raise ValueError("Could not convert to boolean: %s" % str(val))
    return res


def name_range_checker(name, valid_names, descr, replacements=None):
    name = name.strip().lower()
    if replacements is not None and name in replacements:
        name = replacements[name]
    valid = name in valid_names
    assert valid, "Invalid %s '%s'. Available are %s" % (descr, name, str(valid_names))
    return name


# ------------------------------------------------------------------------------
# ---------------------------- Validators --------------------------------------
# ------------------------------------------------------------------------------


@validator
def optional_string_validator(val):
    if len(val.strip()) == 0:
        return None
    return val


@validator
def file_name_validator(name):
    assert len(name) >= 1, "Name should be non-empty"
    return name


@validator
def file_location_validator(location):
    assert path.isfile(location), "location must be a file"
    return os.path.abspath(location)


@validator
def optional_file_location_validator(location):
    if len(location.strip()) > 0:
        assert path.isfile(location), "location must be a file"
        return os.path.abspath(location)
    return None


@validator
def optional_values_file_validator(location):
    if len(location.strip()) == 0:
        return None
    if path.splitext(location)[-1].strip() == "":
        # Assume path to h5 dataset. Validation is done later.
        if "://" not in location:
            location = "silx://" + os.path.abspath(location)
    else:
        # Assume plaintext file
        assert path.isfile(location), "Invalid file path"
        location = os.path.abspath(location)
    return location


@validator
def directory_location_validator(location):
    assert path.isdir(location), "location must be a directory"
    return os.path.abspath(location)


@validator
def optional_directory_location_validator(location):
    if len(location.strip()) > 0:
        assert is_writeable(location), "Directory must be writeable"
        return os.path.abspath(location)
    return None


@validator
def dataset_location_validator(location):
    if not (path.isdir(location)):
        assert (
            path.isfile(location) and path.splitext(location)[-1].split(".")[-1].lower() in files_formats
        ), "Dataset location must be a directory or a HDF5 file"
    return os.path.abspath(location)


@validator
def directory_writeable_validator(location):
    assert is_writeable(location), "Directory must be writeable"
    return os.path.abspath(location)


@validator
def optional_output_directory_validator(location):
    if len(location.strip()) > 0:
        return directory_writeable_validator(location)
    return None


@validator
def optional_output_file_path_validator(location):
    if len(location.strip()) > 0:
        dirname, fname = path.split(location)
        assert os.access(dirname, os.W_OK), "Directory must be writeable"
        return os.path.abspath(location)
    return None


@validator
def integer_validator(val):
    val_int, error = convert_to_int(val)
    assert error is None, "number must be an integer"
    return val_int


@validator
def nonnegative_integer_validator(val):
    val_int, error = convert_to_int(val)
    assert error is None and val_int >= 0, "number must be a non-negative integer"
    return val_int


@validator
def positive_integer_validator(val):
    val_int, error = convert_to_int(val)
    assert error is None and val_int > 0, "number must be a positive integer"
    return val_int


@validator
def optional_positive_integer_validator(val):
    if len(val.strip()) == 0:
        return None
    val_int, error = convert_to_int(val)
    assert error is None and val_int > 0, "number must be a positive integer"
    return val_int


@validator
def nonzero_integer_validator(val):
    val_int, error = convert_to_int(val)
    assert error is None and val_int != 0, "number must be a non-zero integer"
    return val_int


@validator
def binning_validator(val):
    if val == "":
        val = "1"
    val_int, error = convert_to_int(val)
    assert error is None and val_int >= 0, "number must be a non-negative integer"
    return max(1, val_int)


@validator
def projections_subsampling_validator(val):
    val = val.strip()
    err_msg = "projections_subsampling: expected one positive integer or two integers in the format step:begin"
    if ":" not in val:
        val += ":0"
    step, begin = val.split(":")
    step_int, error1 = convert_to_int(step)
    begin_int, error2 = convert_to_int(begin)
    if error1 is not None or error2 is not None or step_int <= 0 or begin_int < 0:
        raise ValueError(err_msg)
    return step_int, begin_int


@validator
def optional_file_name_validator(val):
    if len(val) > 0:
        assert len(val) >= 1, "Name should be non-empty"
        assert path.basename(val) == val, "File name should not be a path (no '/')"
        return val
    return None


@validator
def boolean_validator(val):
    res, error = convert_to_bool(val)
    assert error is None, "Invalid boolean value"
    return res


@validator
def boolean_or_auto_validator(val):
    res, error = convert_to_bool(val)
    if error is not None:
        assert val.lower() == "auto", "Valid values are 0, 1 and auto"
        return val
    return res


@validator
def float_validator(val):
    val_float, error = convert_to_float(val)
    assert error is None, "Invalid number"
    return val_float


@validator
def optional_float_validator(val):
    if isinstance(val, float):
        return val
    elif len(val.strip()) >= 1:
        val_float, error = convert_to_float(val)
        assert error is None, "Invalid number"
    else:
        val_float = None
    return val_float


@validator
def optional_nonzero_float_validator(val):
    if isinstance(val, float):
        val_float = val
    elif len(val.strip()) >= 1:
        val_float, error = convert_to_float(val)
        assert error is None, "Invalid number"
    else:
        val_float = None
    if val_float is not None:
        if abs(val_float) < 1e-6:
            val_float = None
    return val_float


@validator
def optional_tuple_of_floats_validator(val):
    if len(val.strip()) == 0:
        return None
    err_msg = "Expected a tuple of two numbers, but got %s" % val
    try:
        res = tuple(float(x) for x in val.strip("()").split(","))
    except Exception as exc:
        raise ValueError(err_msg)
    if len(res) != 2:
        raise ValueError(err_msg)
    return res


@validator
def cor_validator(val):
    val_float, error = convert_to_float(val)
    if error is None:
        return val_float
    if len(val.strip()) == 0:
        return None
    val = name_range_checker(
        val.lower(), set(cor_methods.values()), "center of rotation estimation method", replacements=cor_methods
    )
    return val


@validator
def tilt_validator(val):
    val_float, error = convert_to_float(val)
    if error is None:
        return val_float
    if len(val.strip()) == 0:
        return None
    val = name_range_checker(
        val.lower(), set(tilt_methods.values()), "automatic detector tilt estimation method", replacements=tilt_methods
    )
    return val


@validator
def slice_num_validator(val):
    val_int, error = convert_to_int(val)
    if error is None:
        return val_int
    else:
        assert val in [
            "first",
            "middle",
            "last",
        ], "Expected start_z and end_z to be either a number or first, middle or last"
        return val


@validator
def generic_options_validator(val):
    if len(val.strip()) == 0:
        return None
    return val


cor_options_validator = generic_options_validator


@validator
def cor_slice_validator(val):
    if len(val) == 0:
        return None
    val_int, error = convert_to_int(val)
    if error:
        supported = ["top", "first", "bottom", "last", "middle"]
        assert val in supported, "Invalid value, must be a number or one of %s" % supported
        return val
    else:
        return val_int


@validator
def flatfield_enabled_validator(val):
    return name_range_checker(val, set(flatfield_modes.values()), "flatfield mode", replacements=flatfield_modes)


@validator
def phase_method_validator(val):
    return name_range_checker(
        val, set(phase_retrieval_methods.values()), "phase retrieval method", replacements=phase_retrieval_methods
    )


@validator
def detector_distortion_correction_validator(val):
    return name_range_checker(
        val,
        set(detector_distortion_correction_methods.values()),
        "detector_distortion_correction_methods",
        replacements=detector_distortion_correction_methods,
    )


@validator
def unsharp_method_validator(val):
    return name_range_checker(
        val, set(unsharp_methods.values()), "unsharp mask method", replacements=phase_retrieval_methods
    )


@validator
def padding_mode_validator(val):
    return name_range_checker(val, set(padding_modes.values()), "padding mode", replacements=padding_modes)


@validator
def reconstruction_method_validator(val):
    return name_range_checker(
        val, set(reconstruction_methods.values()), "reconstruction method", replacements=reconstruction_methods
    )


@validator
def fbp_filter_name_validator(val):
    return name_range_checker(
        val,
        set(fbp_filters.values()),
        "FBP filter",
        replacements=fbp_filters,
    )


@validator
def reconstruction_implementation_validator(val):
    return name_range_checker(
        val,
        set(reco_implementations.values()),
        "Reconstruction method implementation",
        replacements=reco_implementations,
    )


@validator
def optimization_algorithm_name_validator(val):
    return name_range_checker(
        val, set(optim_algorithms.values()), "optimization algorithm name", replacements=iterative_methods
    )


@validator
def output_file_format_validator(val):
    return name_range_checker(val, set(files_formats.values()), "output file format", replacements=files_formats)


@validator
def distribution_method_validator(val):
    val = name_range_checker(
        val, set(distribution_methods.values()), "workload distribution method", replacements=distribution_methods
    )
    # TEMP.
    if val != "local":
        raise NotImplementedError("Computation method '%s' is not implemented yet" % val)
    # --
    return val


@validator
def sino_normalization_validator(val):
    val = name_range_checker(
        val, set(sino_normalizations.values()), "sinogram normalization method", replacements=sino_normalizations
    )
    return val


@validator
def sino_deringer_methods(val):
    val = name_range_checker(
        val,
        set(rings_methods.values()),
        "sinogram rings artefacts correction method",
        replacements=rings_methods,
    )
    return val


@validator
def list_of_int_validator(val):
    ids = val.replace(",", " ").split()
    res = list(map(convert_to_int, ids))
    err = list(filter(lambda x: x[1] is not None or x[0] < 0, res))
    if err != []:
        raise ValueError("Could not convert to a list of GPU IDs: %s" % val)
    return list(set(map(lambda x: x[0], res)))


@validator
def list_of_shift_validator(values):
    ids = values.replace(" ", "").split(",")
    return [int(val) if val not in ("auto", "'auto'", '"auto"') else "auto" for val in ids]


@validator
def list_of_tomoscan_identifier(val):
    # TODO: insure those are valid tomoscan identifier
    return val


@validator
def resources_validator(val):
    val = val.strip()
    is_percentage = False
    if "%" in val:
        is_percentage = True
        val = val.replace("%", "")
    val_float, conversion_error = convert_to_float(val)
    assert conversion_error is None, str("Error while converting %s to float" % val)
    return (val_float, is_percentage)


@validator
def walltime_validator(val):
    # HH:mm:ss
    vals = val.strip().split(":")
    error_msg = "Invalid walltime format, expected HH:mm:ss"
    assert len(vals) == 3, error_msg
    hours, mins, secs = vals
    hours, err1 = convert_to_int(hours)
    mins, err2 = convert_to_int(mins)
    secs, err3 = convert_to_int(secs)
    assert err1 is None and err2 is None and err3 is None, error_msg
    err = hours < 0 or mins < 0 or mins > 59 or secs < 0 or secs > 59
    assert err is False, error_msg
    return hours, mins, secs


@validator
def nonempty_string_validator(val):
    assert val != "", "Value cannot be empty"
    return val


@validator
def logging_validator(val):
    return name_range_checker(val, set(log_levels.values()), "logging level", replacements=log_levels)


@validator
def exclude_projections_validator(val):
    val = val.strip()
    if val == "":
        return None
    if path.isfile(val):
        # previous/default behavior
        return {"type": "indices", "file": val}
    if "=" not in val:
        raise ValueError(
            "exclude_projections: expected either 'angles=angles_file.txt' or 'indices=indices_file.txt' or 'angular_range=[a,b]'"
        )
    excl_type, excl_val = val.split("=")
    excl_type = excl_type.strip()
    excl_val = excl_val.strip()
    check_supported(excl_type, exclude_projections_type.keys(), "exclude_projections type")
    if excl_type == "angular_range":

        def _get_range(range_val):
            for c in ["(", ")", "[", "]"]:
                range_val = range_val.replace(c, "")
            r_min, r_max = range_val.split(",")
            return (float(r_min), float(r_max))

        return {"type": "angular_range", "range": _get_range(excl_val)}
    else:
        return {"type": excl_type, "file": excl_val}


@validator
def no_validator(val):
    return val



import numpy as np
from scipy.fft import rfft, irfft
from silx.image.tomography import get_next_power
from ..thirdparty.pore3d_deringer_munch import munchetal_filter
from ..utils import get_2D_3D_shape, get_num_threads, check_supported
from ..misc.fourier_filters import get_bandpass_filter

try:
    from algotom.prep.removal import remove_all_stripe

    __has_algotom__ = True
except ImportError:
    __has_algotom__ = False


class MunchDeringer:
    def __init__(self, sigma, sinos_shape, levels=None, wname="db15", padding=None, padding_mode="edge"):
        """
        Initialize a "Munch Et Al" sinogram deringer. See References for more information.

        Parameters
        -----------
        sigma: float
            Standard deviation of the damping parameter. The higher value of sigma,
            the more important the filtering effect on the rings.
        levels: int, optional
            Number of wavelets decomposition levels.
            By default (None), the maximum number of decomposition levels is used.
        wname: str, optional
            Default is "db15" (Daubechies, 15 vanishing moments)
        sinos_shape: tuple, optional
            Shape of the sinogram (or sinograms stack).
        padding: tuple of two int, optional
            Horizontal padding to use for reducing the aliasing artefacts

        References
        ----------
        B. Munch, P. Trtik, F. Marone, M. Stampanoni, Stripe and ring artifact removal with
        combined wavelet-Fourier filtering, Optics Express 17(10):8567-8591, 2009.
        """
        self._get_shapes(sinos_shape, padding)
        self.sigma = sigma
        self.levels = levels
        self.wname = wname
        self.padding_mode = padding_mode
        self._check_can_use_wavelets()

    def _get_shapes(self, sinos_shape, padding):
        n_z, n_a, n_x = get_2D_3D_shape(sinos_shape)
        self.sinos_shape = n_z, n_a, n_x
        self.n_angles = n_a
        self.n_z = n_z
        self.n_x = n_x
        # Handle "padding=True" or "padding=False"
        if isinstance(padding, bool):
            if padding:
                padding = (n_x // 2, n_x // 2)
            else:
                padding = None
        #
        if padding is not None:
            pad_x1, pad_x2 = padding
            if np.iterable(pad_x1) or np.iterable(pad_x2):
                raise ValueError("Expected padding in the form (x1, x2)")
            self.sino_padded_shape = (n_a, n_x + pad_x1 + pad_x2)
        self.padding = padding

    def _check_can_use_wavelets(self):
        if munchetal_filter is None:
            raise ValueError("Need pywavelets to use this class")

    def _destripe_2D(self, sino, output):
        if self.padding is not None:
            sino = np.pad(sino, ((0, 0), self.padding), mode=self.padding_mode)
        res = munchetal_filter(sino, self.levels, self.sigma, wname=self.wname)
        if self.padding is not None:
            res = res[:, self.padding[0] : -self.padding[1]]
        output[:] = res
        return output

    def remove_rings(self, sinos, output=None):
        """
        Main function to performs rings artefacts removal on sinogram(s).
        CAUTION: this function defaults to in-place processing, meaning that
        the sinogram(s) you pass will be overwritten.

        Parameters
        ----------
        sinos: numpy.ndarray
            Sinogram or stack of sinograms.
        output: numpy.ndarray, optional
            Output array. If set to None (default), the output overwrites the input.
        """
        if output is None:
            output = sinos
        if sinos.ndim == 2:
            return self._destripe_2D(sinos, output)
        n_sinos = sinos.shape[0]
        for i in range(n_sinos):
            self._destripe_2D(sinos[i], output[i])
        return output


class VoDeringer:
    """
    An interface to Nghia Vo's "remove_all_stripe".
    Needs algotom to run.
    """

    def __init__(self, sinos_shape, **remove_all_stripe_options):
        self._check_requirement()
        self._get_shapes(sinos_shape)
        self._remove_all_stripe_kwargs = remove_all_stripe_options

    def _check_requirement(self):
        if not __has_algotom__:
            raise ImportError("Need algotom")

    def _get_shapes(self, sinos_shape):
        n_z, n_a, n_x = get_2D_3D_shape(sinos_shape)
        self.sinos_shape = n_z, n_a, n_x
        self.n_angles = n_a
        self.n_z = n_z
        self.n_x = n_x

    def remove_rings_sinogram(self, sino, output=None):
        new_sino = remove_all_stripe(sino, **self._remove_all_stripe_kwargs)  # out-of-place
        if output is not None:
            output[:] = new_sino[:]
            return output
        return new_sino

    def remove_rings_sinograms(self, sinos, output=None):
        if output is None:
            output = sinos
        for i in range(sinos.shape[0]):
            output[i] = self.remove_rings_sinogram(sinos[i])
        return output

    def remove_rings_radios(self, radios):
        sinos = np.moveaxis(radios, 1, 0)  # (n_a, n_z, n_x) --> (n_z, n_a, n_x)
        return self.remove_rings_sinograms(sinos)

    remove_rings = remove_rings_sinograms


class SinoMeanDeringer:
    supported_modes = ["subtract", "divide"]

    def __init__(self, sinos_shape, mode="subtract", filter_cutoff=None, padding_mode="edge", fft_num_threads=None):
        """
        Rings correction with mean subtraction/division.
        The principle of this method is to subtract (or divide) the sinogram by its mean along a certain axis.
        In short:
          sinogram -= filt(sinogram.mean(axis=0))
        where `filt` is some bandpass filter.

        Parameters
        ----------
        sinos_shape: tuple of int
            Sinograms shape, in the form (n_angles, n_x) or (n_sinos, n_angles, n_x)
        mode: str, optional
            Operation to do on the sinogram, either "subtract" or "divide"
        filter_cutoff: tuple, optional
            Cut-off of the bandpass filter applied on the sinogram profiles.
            Empty (default) means no filtering.
            Possible values forms are:
              - (sigma_low, sigma_high): two float values defining the standard deviation of
                gaussian(sigma_low) * (1 - gaussian(sigma_high)).
                High values of sigma mean stronger effect of associated filters.
              - ((cutoff_low, transition_low), (cutoff_high, transition_high))
                where "cutoff" is in normalized Nyquist frequency (0.5 is the maximum frequency),
                and "transition" is the width of filter decay in fraction of the cutoff frequency
        padding_mode: str, optional
            Padding mode when filtering the sinogram profile.
            Should be "constant" (i.e "zeros") for mathematical correctness,
            but in practice this yields a Gibbs effect when replicating the sinogram, so "edges" is recommended.
        fft_num_threads: int, optional
            How many threads to use for computing the fast Fourier transform when filtering the sinogram profile.
            Defaut is all the available threads.
        """
        self._get_shapes(sinos_shape)
        check_supported(mode, self.supported_modes, "operation mode")
        self.mode = mode
        self._init_filter(filter_cutoff, fft_num_threads, padding_mode)

    def _get_shapes(self, sinos_shape):
        n_z, n_a, n_x = get_2D_3D_shape(sinos_shape)
        self.sinos_shape = n_z, n_a, n_x
        self.n_angles = n_a
        self.n_z = n_z
        self.n_x = n_x

    def _init_filter(self, filter_cutoff, fft_num_threads, padding_mode):
        self.filter_cutoff = filter_cutoff
        self._filter_f = None
        if filter_cutoff is None:
            return
        self._filter_size = get_next_power(self.n_x * 2)
        self._filter_f = get_bandpass_filter(
            (1, self._filter_size),
            cutoff_lowpass=filter_cutoff[0],
            cutoff_highpass=filter_cutoff[1],
            use_rfft=True,
            data_type=np.float32,
        ).ravel()
        self._fft_n_threads = get_num_threads(fft_num_threads)
        # compat
        if padding_mode == "edges":
            padding_mode = "edge"
        #
        self.padding_mode = padding_mode
        size_diff = self._filter_size - self.n_x
        self._pad_left, self._pad_right = size_diff // 2, size_diff - size_diff // 2

    def _apply_filter(self, sino_profile):
        if self._filter_f is None:
            return sino_profile

        sino_profile = np.pad(sino_profile, (self._pad_left, self._pad_right), mode=self.padding_mode)

        sino_f = rfft(sino_profile, workers=self._fft_n_threads)
        sino_f *= self._filter_f

        return irfft(sino_f, workers=self._fft_n_threads)[self._pad_left : -self._pad_right]  # ascontiguousarray ?

    def remove_rings_sinogram(self, sino, output=None):
        #
        if output is not None:
            raise NotImplementedError
        #
        sino_profile = sino.mean(axis=0)
        sino_profile = self._apply_filter(sino_profile)
        if self.mode == "subtract":
            sino -= sino_profile
        elif self.mode == "divide":
            sino /= sino_profile
        return sino

    def remove_rings_sinograms(self, sinos, output=None):
        #
        if output is not None:
            raise NotImplementedError
        #
        for i in range(sinos.shape[0]):
            self.remove_rings_sinogram(sinos[i])

    remove_rings = remove_rings_sinograms



from copy import copy
from typing import Union
from nabu.stitching.config import SingleAxisStitchingConfiguration
from tomoscan.esrf import NXtomoScan
from tomoscan.volumebase import VolumeBase
from tomoscan.identifier import BaseIdentifier


def get_obj_constant_side_length(obj: Union[NXtomoScan, VolumeBase], axis: int) -> int:
    """
    return tomo object lenght that will be constant over 1D stitching.
    In the case of a stitching along axis 0 this will be:
        * the projection width for pre-processing
        * volume.shape[2] for post-processing

    In the case of a stitching along axis 1 this will be:
        * the projection height for pre-processing
    """
    if isinstance(obj, NXtomoScan):
        if axis == 0:
            return obj.dim_1
        elif axis in (1, 2):
            return obj.dim_2
    elif isinstance(obj, VolumeBase) and axis == 0:
        return obj.get_volume_shape()[-1]
    else:
        raise TypeError(f"obj type ({type(obj)}) and axis == {axis} is not handled")


class _StitcherBase:
    """
    Any stitcher base class
    """

    def __init__(self, configuration, progress=None) -> None:
        if not isinstance(configuration, SingleAxisStitchingConfiguration):
            raise TypeError

        # flag to check if the serie has been ordered yet or not
        self._configuration = copy(configuration)
        # copy configuration because we will edit it
        self._frame_composition = None
        self._progress = progress
        self._overlap_kernels = []
        # kernels to create the stitching on overlaps.

    @property
    def serie_label(self) -> str:
        """return serie name for logs"""
        raise NotImplementedError("Base class")

    @property
    def reading_orders(self):
        """
        as scan can be take on one direction or the order (rotation goes from X to Y then from Y to X)
        we might need to read data from one direction or another
        """
        return self._reading_orders

    def order_input_tomo_objects(self):
        """
        order inputs tomo objects
        """
        raise NotImplementedError("Base class")

    def check_inputs(self):
        """
        order inputs tomo objects
        """
        raise NotImplementedError("Base class")

    def pre_processing_computation(self):
        """
        some specific pre-processing that can be call before retrieving the data
        """
        pass

    @staticmethod
    def param_is_auto(param):
        return param in ("auto", ("auto",))

    def stitch(self, store_composition: bool = True) -> BaseIdentifier:
        """
        Apply expected stitch from configuration and return the DataUrl of the object created

        :param bool store_composition: if True then store the composition used for stitching in frame_composition.
                                       So it can be reused by third part (like tomwer) to display composition made
        """
        raise NotImplementedError("base class")

    @property
    def frame_composition(self):
        return self._frame_composition

    @staticmethod
    def from_abs_pos_to_rel_pos(abs_position: tuple):
        """
        return relative position from on object to the other but in relative this time
        :param tuple abs_position: tuple containing the absolute positions
        :return: len(abs_position) - 1 relative position
        :rtype: tuple
        """
        return tuple([pos_obj_b - pos_obj_a for (pos_obj_a, pos_obj_b) in zip(abs_position[:-1], abs_position[1:])])

    @staticmethod
    def from_rel_pos_to_abs_pos(rel_positions: tuple, init_pos: int):
        """
        return absolute positions from a tuple of relative position and an initial position
        :param tuple rel_positions: tuple containing the absolute positions
        :return: len(rel_positions) + 1 relative position
        :rtype: tuple
        """
        abs_pos = [
            init_pos,
        ]
        for rel_pos in rel_positions:
            abs_pos.append(abs_pos[-1] + rel_pos)
        return abs_pos

    def _compute_shifts(self):
        """
        after this stage the final shifts must be determine
        """
        raise NotImplementedError("base class")



from ..fullfield.nabu_config import *
import copy

## keep the text below for future inclusion in the documentation
# start_z, end_z , start_z_mm, end_z_mm
# ----------------------------------------------------------------
# By default, all the volume is reconstructed slice by slice, along the axis 'z'.
# ** option 1) you can set
#     start_z_mm=0
#     end_z_mm  = 0
#
#     Now, concerning start_z and end_z,  Use positive integers, with start_z < end_z
#               The reconstructed vertical region will be
#                   slice start = first doable + start_z
#                   slice end   = first doable + end_z
#                or less if such range needs to be clipped to the doable one
#
#                As an example start_z= 10, end_z = 20
#                for reconstructing 10 slices close to scan start.
#                NOTE: we are proceeding in the direction of the scan so that, in millimiters,
#                       the start may be above or below the end
#      To reconstruct the whole doable volume set
#
#        start_z= 0
#        end_z  =-1
#
# ** option 2) using start_z_mm, end_z_mm
#               Use positive floats, in millimiters. They indicate the height above the sample stage
#             The values of start_z and end_z are not used in this case


help_start_end_z = """ If  start_z_mm , end_z_mm are seto to zero, then start_z and end_z will be effective unless  end_z_fract  is different from zero. In this latter case the vertical range will be given in terms o the fractional position between the first doable and last doable slices.
Otherwhise, if start_z_mm and end_z_mm are not zero,  the slices whose height above the sample stage, in millimiters, between start_z_mm and end_z_mm are reconstructed
"""

# we need to deepcopy this in order not to mess the original nabu_config of the full-field pipeline
nabu_config = copy.deepcopy(nabu_config)

nabu_config["preproc"]["processes_file"] = {
    "default": "",
    "help": "Path tgo the file where some operations should be stored for later use. By default it is 'xxx_nabu_processes.h5'",
    "validator": optional_file_location_validator,
    "type": "required",
}
nabu_config["preproc"]["double_flatfield_enabled"]["default"] = 1


nabu_config["reconstruction"].update(
    {
        "dz_per_proj": {
            "default": 0,
            "help": " A positive DZPERPROJ means that the rotation axis is going up. Alternatively the vertical translations, can be given through an array using the variable z_per_proj_file",
            "validator": float_validator,
            "type": "optional",
        },
        "z_per_proj_file": {
            "default": "",
            "help": "Alternative to dz_per_proj. A file where each line has one value: vertical displacements of the axis. There should be as many values as there are projection images.",
            "validator": optional_file_location_validator,
            "type": "optional",
        },
        "dx_per_proj": {
            "default": 0,
            "help": " A positive value means that the rotation axis is going on the rigth. Alternatively the horizontal translations, can be given through an array using the variable x_per_proj_file",
            "validator": float_validator,
            "type": "optional",
        },
        "x_per_proj_file": {
            "default": "",
            "help": "Alternative to dx_per_proj. A file where each line has one value: horizontal displacements of the axis. There should be as many values as there are projection images.",
            "validator": optional_file_location_validator,
            "type": "optional",
        },
        "axis_to_the_center": {
            "default": "1",
            "help": "Whether to shift start_x and start_y so to have the axis at the center",
            "validator": boolean_validator,
            "type": "optional",
        },
        "auto_size": {
            "default": "1",
            "help": "Wether to set automatically start_x end_x start_y end_y ",
            "validator": boolean_validator,
            "type": "optional",
        },
        "use_hbp": {
            "default": "0",
            "help": "Wether to use hbp routine instead of the backprojector from fbp ",
            "validator": boolean_validator,
            "type": "optional",
        },
        "fan_source_distance_meters": {
            "default": 1.0e9,
            "help": "For HBP, for the description of the fan geometry, the source to axis distance. Defaults to a large value which implies parallel geometry",
            "validator": float_validator,
            "type": "optional",
        },
        "start_z_mm": {
            "default": "0",
            "help": help_start_end_z,
            "validator": float_validator,
            "type": "optional",
        },
        "end_z_mm": {
            "default": "0",
            "help": " To determine the reconstructed vertical range: the height in millimiters above the stage below which slices are reconstructed ",
            "validator": float_validator,
            "type": "optional",
        },
        "start_z_fract": {
            "default": "0",
            "help": help_start_end_z,
            "validator": float_validator,
            "type": "optional",
        },
        "end_z_fract": {
            "default": "0",
            "help": " To determine the reconstructed vertical range: the height in fractional position between first doable slice and last doable slice  above the stage below which slices are reconstructed ",
            "validator": float_validator,
            "type": "optional",
        },
        "start_z": {
            "default": "0",
            "help": "the first slice of the reconstructed range. Numbered going in the direction of the scan and starting with number zero for the first doable slice",
            "validator": slice_num_validator,
            "type": "optional",
        },
        "end_z": {
            "default": "-1",
            "help": "the "
            "end"
            " slice of the reconstructed range. Numbered going in the direction of the scan and starting with number zero for the first doable slice",
            "validator": slice_num_validator,
            "type": "optional",
        },
    }
)
nabu_config["pipeline"].update(
    {
        "skip_after_flatfield_dump": {
            "default": "0",
            "help": "When the writing of the flatfielded data is activated, if this option is set, then the phase and reconstruction steps are skipped",
            "validator": boolean_validator,
            "type": "optional",
        },
    }
)
nabu_config["reconstruction"].update(
    {
        "angular_tolerance_steps": {
            "default": "3.0",
            "help": "the angular tolerance, an angular width expressed in units of an angular step, which is tolerated in the criteria for deciding if a slice is reconstructable or not",
            "validator": float_validator,
            "type": "advanced",
        },
        "redundancy_angle_deg": {
            "default": "0",
            "help": "Can be 0,180 or 360. If there are dead detector regions (notably scintillator junction (stripes) which need to be complemented at +-360 for local tomo or +- 180 for conventional tomo. This may have an impact on the doable vertical span (you can check it with the --dry-run 1 option)",
            "validator": float_validator,
            "type": "advanced",
        },
        "enable_halftomo": {
            "default": "0",
            "help": "nabu-helical applies the same treatment for half-tomo as for full-tomo. Always let this key to zero",
            "validator": boolean_validator,
            "type": "advanced",
        },
        "helical_halftomo": {
            "default": "1",
            "help": "Wether to consider doable slices those which are contributed by an angular span greater or equal to 360, instead of just 180 or more",
            "validator": boolean_validator,
            "type": "advanced",
        },
    }
)



from math import ceil
from silx.image.tomography import get_next_power
from ...utils import check_supported


def estimate_required_memory(
    process_config, delta_z=None, delta_a=None, max_mem_allocation_GB=None, fft_plans=True, debug=False
):
    """
    Estimate the memory (RAM) in Bytes needed for a reconstruction.

    Parameters
    -----------
    process_config: `ProcessConfig` object
        Data structure with the processing configuration
    delta_z: int, optional
        How many lines are to be loaded in the each projection image.
        Default is to load lines [start_z:end_z] (where start_z and end_z come from the user configuration file)
    delta_a: int, optional
        How many (partial) projection images to load at the same time. Default is to load all the projection images.
    max_mem_allocation_GB: float, optional
        Maximum amount of memory in GB for one single array.

    Returns
    -------
    required_memory: float
        Total required memory (in bytes).

    Raises
    ------
    ValueError if one single-array allocation exceeds "max_mem_allocation_GB"

    Notes
    -----
    pycuda <= 2022.1 cannot use arrays with more than 2**32 items (i.e 17.18 GB for float32).
    This was solved in more recent versions.
    """

    def check_memory_limit(mem_GB, name):
        if max_mem_allocation_GB is None:
            return
        if mem_GB > max_mem_allocation_GB:
            raise ValueError(
                "Cannot allocate array '%s' %.3f GB > max_mem_allocation_GB = %.3f GB"
                % (name, mem_GB, max_mem_allocation_GB)
            )

    dataset = process_config.dataset_info
    processing_steps = process_config.processing_steps

    # The "x" dimension (last axis) is always the image width, because
    #  - Processing is never "cut" along this axis (we either split along frames, or images lines)
    #  - Even if we want to reconstruct a vertical slice (i.e end_x - start_x == 1),
    #    the tomography reconstruction will need the full information along this axis.
    # The only case where reading a x-subregion would be useful is to crop the initial data
    # (i.e knowing that a part of each image will be completely unused). This is not supported yet.
    Nx = process_config.radio_shape(binning=True)[-1]

    if delta_z is not None:
        Nz = delta_z // process_config.binning_z
    else:
        Nz = process_config.rec_delta_z  # accounting for binning
    if delta_a is not None:
        Na = ceil(delta_a / process_config.subsampling_factor)
    else:
        Na = process_config.n_angles(subsampling=True)

    total_memory_needed = 0

    # Read data
    # ----------
    data_image_size = Nx * Nz * 4
    data_volume_size = Na * data_image_size
    check_memory_limit(data_volume_size / 1e9, "projections")
    total_memory_needed += data_volume_size

    # CCD processing
    # ---------------
    if "flatfield" in processing_steps:
        # Flat-field is done in-place, but still need to load darks/flats
        n_darks = len(dataset.darks)
        n_flats = len(dataset.flats)
        total_memory_needed += (n_darks + n_flats) * data_image_size

    if "ccd_correction" in processing_steps:
        # CCD filter is "batched 2D"
        total_memory_needed += data_image_size

    # Phase retrieval
    # ---------------
    if "phase" in processing_steps:
        # Phase retrieval is done image-wise, so near in-place, but needs to allocate some memory:
        # filter with padded shape, radio_padded, radio_padded_fourier, and possibly FFT plan.
        # CTF phase retrieval uses "2 filters" (num and denom) but let's neglect this.
        Nx_p = get_next_power(2 * Nx)
        Nz_p = get_next_power(2 * Nz)
        img_size_real = Nx_p * Nz_p * 4
        img_size_cplx = ((Nx_p * Nz_p) // 2 + 1) * 8  # assuming RFFT
        factor = 1
        if fft_plans:
            factor = 2
        total_memory_needed += (2 * img_size_real + img_size_cplx) * factor

    # Sinogram de-ringing
    # -------------------
    if "sino_rings_correction" in processing_steps:
        method = process_config.processing_options["sino_rings_correction"]["method"]
        if method == "munch":
            # Process is done image-wise.
            # Needs one Discrete Wavelets transform and one FFT/IFFT plan for each scale
            factor = 2 if not (fft_plans) else 5.5  # approx!
            total_memory_needed += (Nx * Na * 4) * factor
        elif method == "vo":
            # cupy-based implementation makes many calls to "scipy-like" functions, where the memory usage is not under control
            # TODO try to estimate this
            pass

    # Reconstruction
    # ---------------
    reconstructed_volume_size = 0
    if "reconstruction" in processing_steps:
        rec_config = process_config.processing_options["reconstruction"]
        Nx_rec = rec_config["end_x"] - rec_config["start_x"] + 1
        Ny_rec = rec_config["end_y"] - rec_config["start_y"] + 1
        reconstructed_volume_size = Nz * Nx_rec * Ny_rec * 4
        check_memory_limit(reconstructed_volume_size / 1e9, "reconstructions")
        total_memory_needed += reconstructed_volume_size
        if process_config.rec_params["method"] == "cone":
            # In cone-beam reconstruction, need both sinograms and reconstruction inside GPU.
            # That's big!
            total_memory_needed += 2 * data_volume_size

    if debug:
        print(
            "Mem for (delta_z=%s, delta_a=%s)  ==>  (Na=%d, Nz=%d, Nx=%d) : %.3f GB"
            % (delta_z, delta_a, Na, Nz, Nx, total_memory_needed / 1e9)
        )

    return total_memory_needed


def estimate_max_chunk_size(
    available_memory_GB,
    process_config,
    pipeline_part="all",
    n_rows=None,
    step=10,
    max_mem_allocation_GB=None,
    fft_plans=True,
    debug=False,
):
    """
    Estimate the maximum size of the data chunk that can be loaded in memory.

    Parameters
    ----------
    available_memory_GB: float
        available memory in Giga Bytes (GB - not GiB !).
    process_config: ProcessConfig
        ProcessConfig object
    pipeline_part: str
        Which pipeline part to consider. Possible options are:
          - "full": Account for all the pipeline steps (reading data all the way to reconstruction).
          - "radios": Consider only the processing steps on projection images (ignore sinogram-based steps and reconstruction)
          - "sinogram": Consider only the processing steps related to sinograms and reconstruction
    n_rows: int, optional
        How many lines to load in each projection. Only accounted for pipeline_part="radios".
    step: int, optional
        Step size when doing the iterative memory estimation
    max_mem_allocation_GB: float, optional
        Maximum size (in GB) for one single array.

    Returns
    -------
    n_max: int
        If pipeline_par is "full" or "sinos": return the maximum number of lines that can be loaded
        in all the projections while fitting memory, i.e `data[:, 0:n_max, :]`
        If pipeline_part is "radios", return the maximum number of (partial) images that can be loaded
        while fitting memory, i.e `data[:, zmin:zmax, 0:n_max]`


    Notes
    -----
    pycuda <= 2022.1 cannot use arrays with more than 2**32 items (i.e 17.18 GB for float32).
    This was solved in more recent versions.
    """

    supported_pipeline_parts = ["all", "radios", "sinos"]
    check_supported(pipeline_part, supported_pipeline_parts, "pipeline_part")

    processing_steps_bak = process_config.processing_steps.copy()
    reconstruction_steps = ["sino_rings_correction", "reconstruction"]

    if pipeline_part == "all":
        # load lines from all the projections
        delta_a = None
        delta_z = 0
    if pipeline_part == "radios":
        # order should not matter
        process_config.processing_steps = list(set(process_config.processing_steps) - set(reconstruction_steps))
        # load lines from only a subset of projections
        delta_a = 0
        delta_z = n_rows
    if pipeline_part == "sinos":
        process_config.processing_steps = [
            step for step in process_config.processing_steps if step in reconstruction_steps
        ]
        # load lines from all the projections
        delta_a = None
        delta_z = 0

    mem = 0
    # pylint: disable=E0606, E0601
    last_valid_delta_a = delta_a
    last_valid_delta_z = delta_z
    while True:
        try:
            mem = estimate_required_memory(
                process_config,
                delta_z=delta_z,
                delta_a=delta_a,
                max_mem_allocation_GB=max_mem_allocation_GB,
                fft_plans=fft_plans,
                debug=debug,
            )
        except ValueError:
            # For very big dataset this function might return "0".
            # Either start at 1, or use a smaller step...
            break
        if mem / 1e9 > available_memory_GB:
            break
        if delta_a is not None and delta_a > process_config.n_angles():
            break
        if delta_z is not None and delta_z > process_config.radio_shape()[0]:
            break
        last_valid_delta_a, last_valid_delta_z = delta_a, delta_z
        if pipeline_part == "radios":
            delta_a += step
        else:
            delta_z += step

    process_config.processing_steps = processing_steps_bak

    if pipeline_part != "radios":
        if mem / 1e9 < available_memory_GB:
            res = min(delta_z, process_config.radio_shape()[0])
        else:
            res = last_valid_delta_z
    else:
        if mem / 1e9 < available_memory_GB:
            res = min(delta_a, process_config.n_angles())
        else:
            res = last_valid_delta_a

    # Really not ideal. For very large dataset, "step" should be very small.
    # Otherwise we go from 0 -> OK to 10 -> not OK, and then retain 0...
    if res == 0:
        res = 1
    #

    return res



import numpy as np
import pytest
from nabu.cuda.utils import get_cuda_context, __has_pycuda__
from nabu.opencl.utils import __has_pyopencl__, get_opencl_context
from nabu.testutils import get_data, generate_tests_scenarios, __do_long_tests__
from nabu.processing.transpose import CudaTranspose, OpenCLTranspose

configs = {
    "shape": [(300, 451), (300, 300), (255, 300)],
    "output_is_none": [True, False],
    "dtype_in_out": [(np.float32, np.float32)],
}

if __do_long_tests__:
    configs["dtype_in_out"].extend(
        [(np.float32, np.complex64), (np.complex64, np.complex64), (np.uint8, np.uint16), (np.uint8, np.int32)]
    )

scenarios = generate_tests_scenarios(configs)


@pytest.fixture(scope="class")
def bootstrap(request):
    cls = request.cls
    cls.data = get_data("chelsea.npz")["data"]
    cls.tol = 1e-7
    if __has_pycuda__:
        cls.cu_ctx = get_cuda_context(cleanup_at_exit=False)
    if __has_pyopencl__:
        cls.cl_ctx = get_opencl_context(device_type="all")
    yield
    if __has_pycuda__:
        cls.cu_ctx.pop()


@pytest.mark.usefixtures("bootstrap")
class TestTranspose:
    def _do_test_transpose(self, config, transpose_cls):
        shape = config["shape"]
        dtype = config["dtype_in_out"][0]
        dtype_out = config["dtype_in_out"][1]
        data = np.ascontiguousarray(self.data[: shape[0], : shape[1]], dtype=dtype)

        backend = transpose_cls.backend
        if backend == "opencl" and not (np.iscomplexobj(dtype(1))) and np.iscomplexobj(dtype_out(1)):
            pytest.skip("pyopencl does not support real to complex scalar cast")
        ctx = self.cu_ctx if backend == "cuda" else self.cl_ctx
        backend_options = {"ctx": ctx}
        transpose = transpose_cls(data.shape, dtype, dst_dtype=dtype_out, **backend_options)

        d_data = transpose.processing.allocate_array("data", shape, dtype)
        d_data.set(data)
        if config["output_is_none"]:
            d_out = None
        else:
            d_out = transpose.processing.allocate_array("output", shape[::-1], dtype_out)

        d_res = transpose(d_data, dst=d_out)

        assert (
            np.max(np.abs(d_res.get() - data.T)) == 0
        ), "something wrong with transpose(shape=%s, dtype=%s, dtype_out=%s)" % (shape, dtype, dtype_out)

    @pytest.mark.skipif(not (__has_pycuda__), reason="Need pycuda for this test")
    @pytest.mark.parametrize("config", scenarios)
    def test_cuda_transpose(self, config):
        self._do_test_transpose(config, CudaTranspose)

    @pytest.mark.skipif(not (__has_pyopencl__), reason="Need pyopencl for this test")
    @pytest.mark.parametrize("config", scenarios)
    def test_opencl_transpose(self, config):
        self._do_test_transpose(config, OpenCLTranspose)



import logging
import os
import sys
import numpy as np
import re

from nabu.resources.dataset_analyzer import HDF5DatasetAnalyzer
from nabu.pipeline.estimators import CompositeCOREstimator, estimate_cor
from nabu.resources.nxflatfield import update_dataset_info_flats_darks
from nabu.resources.utils import extract_parameters
from nxtomo.application.nxtomo import NXtomo
from .. import version
from .cli_configs import CompositeCorConfig
from .utils import parse_params_values
from ..utils import DictToObj
import json


class NumpyArrayEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        return json.JSONEncoder.default(self, obj)


def main(user_args=None):
    "Application to extract with the composite cor finder the center of rotation for a scan or a series of scans"

    if user_args is None:
        user_args = sys.argv[1:]

    args_dict = parse_params_values(
        CompositeCorConfig,
        parser_description=main.__doc__,
        program_version="nabu " + version,
        user_args=user_args,
    )
    composite_cor_entry_point(args_dict)
    # here we have been called by the cli. The return value 0 means OK
    return 0


def composite_cor_entry_point(args_dict):
    args = DictToObj(args_dict)

    if len(os.path.dirname(args.filename_template)) == 0:
        # To make sure that other utility routines can succesfully deal with path within the current directory
        args.filename_template = os.path.join(".", args.filename_template)

    args.filename_template = os.path.abspath(args.filename_template)
    if args.first_stage is not None:
        if args.num_of_stages is None:
            args.num_of_stages = 1
        # if the first_stage parameter has been given then
        # we are using numbers to form the names of the files.
        # The filename must containe a XX..X substring which will be replaced
        pattern = re.compile("[X]+")
        ps = pattern.findall(args.filename_template)
        if len(ps) == 0:
            message = f""" You have specified the "first_stage" parameter, with an integer.
            Therefore the "filename_template" parameter is expected to containe a XX..X subsection
            but none was found in the passed parameter which is {args.filename_template}
            """
            raise ValueError(message)
        ls = list(map(len, ps))
        idx = np.argmax(ls)

        args.filename_template = args.filename_template.replace(ps[idx], "{i_stage:" + "0" + str(ls[idx]) + "d}")

    if args.num_of_stages is None:
        # this way it works also in the simple case where
        # only the filename is provided together with the cor options
        num_of_stages = 1
        first_stage = 0
    else:
        num_of_stages = args.num_of_stages
        first_stage = args.first_stage
    cor_list = []
    for iz in range(first_stage, first_stage + num_of_stages):
        if args.num_of_stages is not None:
            nexus_name = args.filename_template.format(i_stage=iz)
        else:
            nexus_name = args.filename_template

        dataset_info = HDF5DatasetAnalyzer(nexus_name, extra_options={"h5_entry": args.entry_name})
        update_dataset_info_flats_darks(dataset_info, flatfield_mode=1)

        ### JL start ###
        # Extract CoR parameters from configuration file
        try:
            cor_options = extract_parameters(args.cor_options, sep=";")
        except Exception as exc:
            msg = "Could not extract parameters from cor_options: %s" % (str(exc))
            raise ValueError(msg)
        ### JL end ###

        # JL start ###
        #: next bit will be done in estimate
        # if "near_pos" not in args.cor_options:
        #    scan = NXtomo()
        #    scan.load(file_path=nexus_name, data_path=args.entry_name)
        #    estimated_near = scan.instrument.detector.x_rotation_axis_pixel_position
        #
        #    cor_options = args.cor_options + f" ; near_pos = {estimated_near} "
        #
        # else:
        #    cor_options = args.cor_options
        ### JL end ###

        cor_finder = CompositeCOREstimator(
            dataset_info,
            oversampling=args.oversampling,
            theta_interval=args.theta_interval,
            n_subsampling_y=args.n_subsampling_y,
            take_log=True,
            spike_threshold=0.04,
            cor_options=cor_options,
            norm_order=1,
        )

        cor_position = cor_finder.find_cor()

        cor_list.append(cor_position)

    cor_list = np.array(cor_list).T

    if args.output_file is not None:
        output_name = args.output_file
    else:
        output_name = os.path.splitext(args.filename_template)[0] + "_cors.txt"

    if output_name.endswith(".json"):
        with open(output_name, "w") as fp:
            json.dump(
                dict(rotation_axis_position=cor_list[0], rotation_axis_position_list=cor_list),
                fp,
                indent=4,
                cls=NumpyArrayEncoder,
            )
    else:
        np.savetxt(
            output_name,
            cor_list,
        )



from math import ceil
from typing import Any, Iterable, Optional, Union, Sized
from dataclasses import dataclass
import numpy
from pyunitsystem.metricsystem import MetricSystem
from nxtomo.paths import nxtomo
from tomoscan.factory import Factory
from tomoscan.identifier import VolumeIdentifier, ScanIdentifier
from tomoscan.esrf.scan.nxtomoscan import NXtomoScan
from ..pipeline.config_validators import (
    boolean_validator,
    convert_to_bool,
)
from ..utils import concatenate_dict, convert_str_to_tuple
from ..io.utils import get_output_volume
from .overlap import OverlapStitchingStrategy
from .utils.utils import ShiftAlgorithm
from .definitions import StitchingType
from .alignment import AlignmentAxis1, AlignmentAxis2
from pyunitsystem.metricsystem import MetricSystem


KEY_IMG_REG_METHOD = "img_reg_method"

KEY_WINDOW_SIZE = "window_size"

KEY_LOW_PASS_FILTER = "low_pass"

KEY_HIGH_PASS_FILTER = "high_pass"

KEY_OVERLAP_SIZE = "overlap_size"

KEY_SIDE = "side"

OUTPUT_SECTION = "output"

INPUTS_SECTION = "inputs"

PRE_PROC_SECTION = "preproc"

POST_PROC_SECTION = "postproc"

INPUT_DATASETS_FIELD = "input_datasets"

INPUT_PIXEL_SIZE_MM = "pixel_size"

INPUT_VOXEL_SIZE_MM = "voxel_size"

STITCHING_SECTION = "stitching"

STITCHING_STRATEGY_FIELD = "stitching_strategy"

STITCHING_TYPE_FIELD = "type"

DATA_FILE_FIELD = "location"

OVERWRITE_RESULTS_FIELD = "overwrite_results"

DATA_PATH_FIELD = "data_path"

AXIS_0_POS_PX = "axis_0_pos_px"

AXIS_1_POS_PX = "axis_1_pos_px"

AXIS_2_POS_PX = "axis_2_pos_px"

AXIS_0_POS_MM = "axis_0_pos_mm"

AXIS_1_POS_MM = "axis_1_pos_mm"

AXIS_2_POS_MM = "axis_2_pos_mm"

AXIS_0_PARAMS = "axis_0_params"

AXIS_1_PARAMS = "axis_1_params"

AXIS_2_PARAMS = "axis_2_params"

FLIP_LR = "fliplr"

FLIP_UD = "flipud"

NEXUS_VERSION_FIELD = "nexus_version"

OUTPUT_DTYPE = "data_type"

OUTPUT_VOLUME = "output_volume"

STITCHING_SLICES = "slices"

CROSS_CORRELATION_SLICE_FIELD = "slice_index_for_correlation"

RESCALE_FRAMES = "rescale_frames"

RESCALE_PARAMS = "rescale_params"

KEY_RESCALE_MIN_PERCENTILES = "rescale_min_percentile"

KEY_RESCALE_MAX_PERCENTILES = "rescale_max_percentile"

ALIGNMENT_AXIS_2_FIELD = "alignment_axis_2"

ALIGNMENT_AXIS_1_FIELD = "alignment_axis_1"

PAD_MODE_FIELD = "pad_mode"

AVOID_DATA_DUPLICATION_FIELD = "avoid_data_duplication"

# SLURM

SLURM_SECTION = "slurm"

SLURM_PARTITION = "partition"

SLURM_MEM = "memory"

SLURM_COR_PER_TASKS = "cpu-per-task"

SLURM_NUMBER_OF_TASKS = "n_tasks"

SLURM_N_JOBS = "n_jobs"

SLURM_OTHER_OPTIONS = "other_options"

SLURM_PREPROCESSING_COMMAND = "python_venv"

SLURM_MODULES_TO_LOADS = "modules"

SLURM_CLEAN_SCRIPTS = "clean_scripts"

# normalization by sample

NORMALIZATION_BY_SAMPLE_SECTION = "normalization_by_sample"

NORMALIZATION_BY_SAMPLE_ACTIVE_FIELD = "active"

NORMALIZATION_BY_SAMPLE_METHOD = "method"

NORMALIZATION_BY_SAMPLE_SIDE = "side"

NORMALIZATION_BY_SAMPLE_MARGIN = "margin"

NORMALIZATION_BY_SAMPLE_WIDTH = "width"

# kernel extra options

STITCHING_KERNELS_EXTRA_PARAMS = "stitching_kernels_extra_params"

KEY_THRESHOLD_FREQUENCY = "threshold_frequency"


CROSS_CORRELATION_METHODS_AXIS_0 = {
    "": "",  # for display
    ShiftAlgorithm.NABU_FFT.value: "will call nabu `find_shift_correlate` function - shift search in fourier space",
    ShiftAlgorithm.SKIMAGE.value: "use scikit image `phase_cross_correlation` function in real space",
    ShiftAlgorithm.NONE.value: "no shift research is done. will only get shift from motor positions",
}

CROSS_CORRELATION_METHODS_AXIS_2 = CROSS_CORRELATION_METHODS_AXIS_0.copy()
CROSS_CORRELATION_METHODS_AXIS_2.update(
    {
        ShiftAlgorithm.CENTERED.value: "a fast and simple auto-CoR method. It only works when the CoR is not far from the middle of the detector. It does not work for half-tomography.",
        ShiftAlgorithm.GLOBAL.value: "a slow but robust auto-CoR.",
        ShiftAlgorithm.GROWING_WINDOW.value: "automatically find the CoR with a sliding-and-growing window. You can tune the option with the parameter 'cor_options'.",
        ShiftAlgorithm.SLIDING_WINDOW.value: "semi-automatically find the CoR with a sliding window. You have to specify on which side the CoR is (left, center, right). Please see the 'cor_options' parameter.",
        ShiftAlgorithm.COMPOSITE_COARSE_TO_FINE.value: "Estimate CoR from composite multi-angle images. Only works for 360 degrees scans.",
        ShiftAlgorithm.SINO_COARSE_TO_FINE.value: "Estimate CoR from sinogram. Only works for 360 degrees scans.",
    }
)

SECTIONS_COMMENTS = {
    STITCHING_SECTION: "section dedicated to stich parameters\n",
    OUTPUT_SECTION: "section dedicated to output parameters\n",
    INPUTS_SECTION: "section dedicated to inputs\n",
    SLURM_SECTION: "section didicated to slurm. If you want to run locally avoid setting 'partition or remove this section'",
    NORMALIZATION_BY_SAMPLE_SECTION: "section dedicated to normalization by a sample. If activate each frame can be normalized by a sample of the frame",
}

DEFAULT_SHIFT_ALG_AXIS_0 = "nabu-fft"
DEFAULT_SHIFT_ALG_AXIS_2 = "sliding-window"

_shift_algs_axis_0 = "\n            + ".join(
    [f"{key}: {value}" for key, value in CROSS_CORRELATION_METHODS_AXIS_0.items()]
)
_shift_algs_axis_2 = "\n            + ".join(
    [f"{key}: {value}" for key, value in CROSS_CORRELATION_METHODS_AXIS_2.items()]
)

HELP_SHIFT_PARAMS = f"""options for shifts algorithms as `key1=value1,key2=value2`. For now valid keys are:
    - {KEY_OVERLAP_SIZE}: size to apply stitching. If not provided will take the largest size possible'.
    - {KEY_IMG_REG_METHOD}: algorithm to use to find overlaps between the different sections. Possible values are \n        * for axis 0: {_shift_algs_axis_0}\n        * and for axis 2: {_shift_algs_axis_2}
    - {KEY_LOW_PASS_FILTER}: low pass filter value for filtering frames before shift research
    - {KEY_HIGH_PASS_FILTER}: high pass filter value for filtering frames before shift research"""


def _str_to_dict(my_str: Union[str, dict]):
    """convert a string as key_1=value_2;key_2=value_2 to a dict"""
    if isinstance(my_str, dict):
        return my_str
    res = {}
    for key_value in filter(None, my_str.split(";")):
        key, value = key_value.split("=")
        res[key] = value
    return res


def _dict_to_str(ddict: dict):
    return ";".join([f"{str(key)}={str(value)}" for key, value in ddict.items()])


def str_to_shifts(my_str: Optional[str]) -> Union[str, tuple]:
    if my_str is None:
        return None
    elif isinstance(my_str, str):
        my_str = my_str.replace(" ", "")
        my_str = my_str.lstrip("[").lstrip("(")
        my_str = my_str.rstrip("]").lstrip(")")
        if my_str == "":
            return None
        try:
            shift = ShiftAlgorithm.from_value(my_str)
        except ValueError:
            shifts_as_str = filter(None, my_str.replace(";", ",").split(","))
            return [float(shift) for shift in shifts_as_str]
        else:
            return shift
    elif isinstance(my_str, (tuple, list)):
        return [float(shift) for shift in my_str]
    else:
        raise TypeError("Only str or tuple of str expected expected")


def _valid_stitching_kernels_params(my_dict: Union[dict, str]):
    if isinstance(my_dict, str):
        my_dict = _str_to_dict(my_str=my_dict)

    valid_keys = (KEY_THRESHOLD_FREQUENCY, KEY_SIDE)
    for key in my_dict.keys():
        if not key in valid_keys:
            raise KeyError(f"{key} is a unrecognized key")
    return my_dict


def _valid_shifts_params(my_dict: Union[dict, str]):
    if isinstance(my_dict, str):
        my_dict = _str_to_dict(my_str=my_dict)

    valid_keys = (
        KEY_WINDOW_SIZE,
        KEY_IMG_REG_METHOD,
        KEY_OVERLAP_SIZE,
        KEY_HIGH_PASS_FILTER,
        KEY_LOW_PASS_FILTER,
        KEY_SIDE,
    )
    for key in my_dict.keys():
        if not key in valid_keys:
            raise KeyError(f"{key} is a unrecognized key")
    return my_dict


def _slices_to_list_or_slice(my_str: Optional[str]) -> Union[str, slice]:
    if my_str is None:
        return None
    if isinstance(my_str, (tuple, list)):
        if len(my_str) == 2:
            return slice(int(my_str[0]), int(my_str[1]))
        elif len(my_str) == 3:
            return slice(int(my_str[0]), int(my_str[1]), int(my_str[2]))
        else:
            raise ValueError("expect at most free values to define a slice")

    assert isinstance(my_str, str), f"wrong type. Get {my_str}, {type(my_str)}"
    my_str = my_str.replace(" ", "")
    if ":" in my_str:
        split_string = my_str.split(":")
        start = int(split_string[0])
        stop = int(split_string[1])
        if len(split_string) == 2:
            step = None
        elif len(split_string) == 3:
            step = int(split_string[2])
        else:
            raise ValueError(f"unable to interpret `slices` parameter: {my_str}")
        return slice(start, stop, step)
    else:
        my_str.replace(",", ";")
        return list(filter(None, my_str.split(";")))


def _scalar_or_tuple_to_bool_or_tuple_of_bool(my_str: Union[bool, tuple, str], default=False):
    if isinstance(my_str, bool):
        return my_str
    elif isinstance(my_str, str):
        my_str = my_str.replace(" ", "")
        my_str = my_str.lstrip("(").lstrip("[")
        my_str = my_str.rstrip(")").lstrip("]")
        my_str = my_str.replace(",", ";")
        values = my_str.split(";")
        values = tuple([convert_to_bool(value)[0] for value in values])
    else:
        values = my_str
    if len(values) == 0:
        return default
    elif len(values) == 1:
        return values[0]
    else:
        return values


from nabu.stitching.sample_normalization import Method, SampleSide


class NormalizationBySample:
    def __init__(self) -> None:
        self._active = False
        self._method = Method.MEAN
        self._margin = 0
        self._side = SampleSide.LEFT
        self._width = 30

    def is_active(self):
        return self._active

    def set_is_active(self, active: bool):
        assert isinstance(
            active, bool
        ), f"active is expected to be a bool. Get {type(active)} instead. Value == {active}"
        self._active = active

    @property
    def method(self) -> Method:
        return self._method

    @method.setter
    def method(self, method: Union[Method, str]) -> None:
        self._method = Method.from_value(method)

    @property
    def margin(self) -> int:
        return self._margin

    @margin.setter
    def margin(self, margin: int):
        assert isinstance(margin, int), f"margin is expected to be an int. Get {type(margin)} instead"
        self._margin = margin

    @property
    def side(self) -> SampleSide:
        return self._side

    @side.setter
    def side(self, side: Union[SampleSide, str]):
        self._side = SampleSide.from_value(side)

    @property
    def width(self) -> int:
        return self._width

    @width.setter
    def width(self, width: int):
        assert isinstance(width, int), f"width is expected to be an int. Get {type(width)} instead"

    @staticmethod
    def from_dict(my_dict: dict):
        sample_normalization = NormalizationBySample()
        # active
        active = my_dict.get(NORMALIZATION_BY_SAMPLE_ACTIVE_FIELD, None)
        if active is not None:
            active = active in (True, "True", 1, "1")
            sample_normalization.set_is_active(active)

        # method
        method = my_dict.get(NORMALIZATION_BY_SAMPLE_METHOD, None)
        if method is not None:
            sample_normalization.method = method

        # margin
        margin = my_dict.get(NORMALIZATION_BY_SAMPLE_MARGIN, None)
        if margin is not None:
            sample_normalization.margin = int(margin)

        # side
        side = my_dict.get(NORMALIZATION_BY_SAMPLE_SIDE, None)
        if side is not None:
            sample_normalization.side = side

        # width
        width = my_dict.get(NORMALIZATION_BY_SAMPLE_WIDTH, None)
        if width is not None:
            sample_normalization.width = int(width)

        return sample_normalization

    def to_dict(self) -> dict:
        return {
            NORMALIZATION_BY_SAMPLE_ACTIVE_FIELD: self.is_active(),
            NORMALIZATION_BY_SAMPLE_METHOD: self.method.value,
            NORMALIZATION_BY_SAMPLE_MARGIN: self.margin,
            NORMALIZATION_BY_SAMPLE_SIDE: self.side.value,
            NORMALIZATION_BY_SAMPLE_WIDTH: self.width,
        }

    def __eq__(self, __value: object) -> bool:
        if not isinstance(__value, NormalizationBySample):
            return False
        else:
            return self.to_dict() == __value.to_dict()


@dataclass
class SlurmConfig:
    "configuration for slurm jobs"
    partition: str = ""  # note: must stay empty to make by default we don't use slurm (use by the  configuration file)
    mem: str = "128"
    n_jobs: int = 1
    other_options: str = ""
    preprocessing_command: str = ""
    modules_to_load: tuple = tuple()
    clean_script: bool = ""
    n_tasks: int = 1
    n_cpu_per_task: int = 4

    def __post_init__(self) -> None:
        # make sure either 'modules' or 'preprocessing_command' is provided
        if len(self.modules_to_load) > 0 and self.preprocessing_command not in (None, ""):
            raise ValueError(
                f"Either modules {SLURM_MODULES_TO_LOADS} or preprocessing_command {SLURM_PREPROCESSING_COMMAND} can be provided. Not both."
            )

    def to_dict(self) -> dict:
        "dump configuration to dict"
        return {
            SLURM_PARTITION: self.partition if self.partition is not None else "",
            SLURM_MEM: self.mem,
            SLURM_N_JOBS: self.n_jobs,
            SLURM_OTHER_OPTIONS: self.other_options,
            SLURM_PREPROCESSING_COMMAND: self.preprocessing_command,
            SLURM_MODULES_TO_LOADS: self.modules_to_load,
            SLURM_CLEAN_SCRIPTS: self.clean_script,
            SLURM_NUMBER_OF_TASKS: self.n_tasks,
            SLURM_COR_PER_TASKS: self.n_cpu_per_task,
        }

    @staticmethod
    def from_dict(config: dict):
        return SlurmConfig(
            partition=config.get(
                SLURM_PARTITION, None
            ),  # warning: never set a default value. Would generate infinite loop from slurm call
            mem=config.get(SLURM_MEM, "32GB"),
            n_jobs=int(config.get(SLURM_N_JOBS, 10)),
            other_options=config.get(SLURM_OTHER_OPTIONS, ""),
            n_tasks=config.get(SLURM_NUMBER_OF_TASKS, 1),
            n_cpu_per_task=config.get(SLURM_COR_PER_TASKS, 4),
            preprocessing_command=config.get(SLURM_PREPROCESSING_COMMAND, ""),
            modules_to_load=convert_str_to_tuple(config.get(SLURM_MODULES_TO_LOADS, "")),
            clean_script=convert_to_bool(config.get(SLURM_CLEAN_SCRIPTS, False))[0],
        )


def _cast_shift_to_str(shifts: Union[tuple, str, None]) -> str:
    if shifts is None:
        return ""
    elif isinstance(shifts, ShiftAlgorithm):
        return shifts.value
    elif isinstance(shifts, str):
        return shifts
    elif isinstance(shifts, (tuple, list)):
        return ";".join([str(value) for value in shifts])


@dataclass
class StitchingConfiguration:
    """
    bass class to define stitching configuration
    """

    axis_0_pos_px: Union[tuple, str, None]
    "position along axis 0 in absolute. unit: px"
    axis_1_pos_px: Union[tuple, str, None]
    "position along axis 1 in absolute. unit: px"
    axis_2_pos_px: Union[tuple, str, None]
    "position along axis 2 in absolute. unit: px"

    axis_0_pos_mm: Union[tuple, str, None] = None
    "position along axis 0 in absolute. unit: mm"
    axis_1_pos_mm: Union[tuple, str, None] = None
    "position along axis 0 in absolute. unit: mm"
    axis_2_pos_mm: Union[tuple, str, None] = None
    "position along axis 0 in absolute. unit: mm"

    axis_0_params: dict = None
    axis_1_params: dict = None
    axis_2_params: dict = None
    slurm_config: SlurmConfig = None
    flip_lr: Union[tuple, bool] = False
    "flip frame left-right. For scan this will be append to the NXtransformations of the detector"
    flip_ud: Union[tuple, bool] = False
    "flip frame up-down. For scan this will be append to the NXtransformations of the detector"

    overwrite_results: bool = False
    stitching_strategy: OverlapStitchingStrategy = OverlapStitchingStrategy.COSINUS_WEIGHTS
    stitching_kernels_extra_params: dict = None

    slice_for_cross_correlation: Union[str, int] = "middle"

    # opts for rescaling frame during stitching
    rescale_frames: bool = False
    rescale_params: dict = None

    normalization_by_sample: NormalizationBySample = None

    duplicate_data: bool = True
    """when possible (for HDF5) avoid duplicating data as-much-much-as-possible. Overlaping region between two frames will be duplicated. Remaining will be 'raw_data' for volume.
    For projection flat field will be applied"""

    @property
    def stitching_type(self):
        raise NotImplementedError("Base class")

    def __post_init__(self):
        if self.normalization_by_sample is None:
            self.normalization_by_sample = NormalizationBySample()

    @staticmethod
    def get_description_dict() -> dict:
        def get_pos_info(axis, unit, alternative):
            return f"position over {axis} in {unit}. If provided {alternative} must be set to blank. If none provided then will try to get information from existing metadata"

        def get_default_shift_params(window_size=None, shift_alg=None) -> str:
            return ";".join(
                [
                    f"{KEY_WINDOW_SIZE}={window_size or ''}",
                    f"{KEY_IMG_REG_METHOD}={shift_alg or ''}",
                ]
            )

        return {
            STITCHING_SECTION: {
                STITCHING_TYPE_FIELD: {
                    "default": StitchingType.Z_PREPROC.value,
                    "help": f"stitching to be applied. Must be in {StitchingType.values()}",
                    "type": "required",
                },
                STITCHING_STRATEGY_FIELD: {
                    "default": "cosinus weights",
                    "help": f"Policy to apply to compute the overlap area. Must be in {OverlapStitchingStrategy.values()}.",
                    "type": "required",
                },
                CROSS_CORRELATION_SLICE_FIELD: {
                    "default": "middle",
                    "help": f"slice to use for image registration",
                    "type": "optional",
                },
                AXIS_0_POS_PX: {
                    "default": "",
                    "help": get_pos_info(axis=0, unit="pixel", alternative=AXIS_0_POS_MM),
                    "type": "optional",
                },
                AXIS_0_POS_MM: {
                    "default": "",
                    "help": get_pos_info(axis=1, unit="millimeter", alternative=AXIS_0_POS_PX),
                    "type": "optional",
                },
                AXIS_0_PARAMS: {
                    "default": get_default_shift_params(window_size=50, shift_alg=DEFAULT_SHIFT_ALG_AXIS_0),
                    "help": HELP_SHIFT_PARAMS,
                    "type": "optional",
                },
                AXIS_1_POS_PX: {
                    "default": "",
                    "help": get_pos_info(axis=1, unit="pixel", alternative=AXIS_1_POS_MM),
                    "type": "optional",
                },
                AXIS_1_POS_MM: {
                    "default": "",
                    "help": get_pos_info(axis=1, unit="millimeter", alternative=AXIS_1_POS_PX),
                    "type": "optional",
                },
                AXIS_1_PARAMS: {
                    "default": get_default_shift_params(),
                    "help": f"same as {AXIS_0_PARAMS} but for axis 1",
                    "type": "optional",
                },
                AXIS_2_POS_PX: {
                    "default": "",
                    "help": get_pos_info(axis=2, unit="pixel", alternative=AXIS_2_POS_MM),
                    "type": "optional",
                },
                AXIS_2_POS_MM: {
                    "default": "",
                    "help": get_pos_info(axis=2, unit="millimeter", alternative=AXIS_1_POS_PX),
                    "type": "optional",
                },
                AXIS_2_PARAMS: {
                    "default": get_default_shift_params(window_size=200, shift_alg=DEFAULT_SHIFT_ALG_AXIS_2),
                    "help": f"same as {AXIS_0_PARAMS} but for axis 2",
                    "type": "optional",
                },
                FLIP_LR: {
                    "default": False,
                    "help": "sometime scan or volume can have a left-right flip in frame (projection/slice) space. For recent NXtomo it should be handled automatically. But for volume you might need to request some flip.",
                    "type": "optional",
                },
                FLIP_UD: {
                    "default": False,
                    "help": "sometime scan or volume can have a up_down flip in frame (projection/slice) space. For recent NXtomo it should be handled automatically. But for volume you might need to request some flip.",
                    "type": "optional",
                },
                RESCALE_FRAMES: {
                    "default": False,
                    "help": "rescale each frame before applying stithcing",
                    "type": "advanced",
                },
                RESCALE_PARAMS: {
                    "default": "",
                    "help": f"parameters for rescaling frames as 'key1=value1;key_2=value2'. Valid Keys are {KEY_RESCALE_MIN_PERCENTILES} and {KEY_RESCALE_MAX_PERCENTILES}.",
                    "type": "advanced",
                },
                STITCHING_KERNELS_EXTRA_PARAMS: {
                    "default": "",
                    "help": f"advanced parameters for some stitching kernels. must be provided as 'key1=value1;key_2=value2'. Valid keys for now are: {KEY_THRESHOLD_FREQUENCY}: threshold to be used by the {OverlapStitchingStrategy.IMAGE_MINIMUM_DIVERGENCE.value} to split images low and high frequencies in Fourier space.",
                    "type": "advanced",
                },
                ALIGNMENT_AXIS_2_FIELD: {
                    "default": "center",
                    "help": f"In case frame have different frame widths how to align them (so along volume axis 2). Valid keys are {AlignmentAxis2.values()}",
                    "type": "advanced",
                },
                PAD_MODE_FIELD: {
                    "default": "constant",
                    "help": f"pad mode to use for frame alignment. Valid values are 'constant', 'edge', 'linear_ramp', maximum', 'mean', 'median', 'minimum', 'reflect', 'symmetric', 'wrap', and 'empty'. See nupy.pad documentation for details",
                    "type": "advanced",
                },
                AVOID_DATA_DUPLICATION_FIELD: {
                    "default": "1",
                    "help": "When possible (stitching on reconstructed volume and HDF5 volume as input and output) create link to original data instead of duplicating it all. Warning: this will create relative link between the stiched volume and the original reconstructed volume.",
                    "validator": boolean_validator,
                    "type": "advanced",
                },
            },
            OUTPUT_SECTION: {
                OVERWRITE_RESULTS_FIELD: {
                    "default": "1",
                    "help": "What to do in the case where the output file exists.\nBy default, the output data is never overwritten and the process is interrupted if the file already exists.\nSet this option to 1 if you want to overwrite the output files.",
                    "validator": boolean_validator,
                    "type": "required",
                },
            },
            INPUTS_SECTION: {
                INPUT_DATASETS_FIELD: {
                    "default": "",
                    "help": f"Dataset to stitch together. Must be volume for {StitchingType.Z_PREPROC.value} or NXtomo for {StitchingType.Z_POSTPROC.value}",
                    "type": "required",
                },
                STITCHING_SLICES: {
                    "default": "",
                    "help": f"slices to be stitched. Must be given along axis 0 for pre-processing (z) and along axis 1 for post-processing (y)",
                    "type": "advanced",
                },
            },
            SLURM_SECTION: {
                SLURM_PARTITION: {
                    "default": "",
                    "help": "slurm partition to be used. If empty will run locally",
                    "type": "optional",
                },
                SLURM_MEM: {
                    "default": "32GB",
                    "help": "memory to allocate for each job",
                    "type": "optional",
                },
                SLURM_N_JOBS: {
                    "default": 10,
                    "help": "number of job to launch (split computation on N parallel jobs). Once all are finished we will concatenate the result.",
                    "type": "optional",
                },
                SLURM_COR_PER_TASKS: {
                    "default": 4,
                    "help": "number of cor per task launched",
                    "type": "optional",
                },
                SLURM_NUMBER_OF_TASKS: {
                    "default": 1,
                    "help": "(for parallel execution when possible). Split each job into this number of tasks",
                    "type": "optional",
                },
                SLURM_OTHER_OPTIONS: {
                    "default": "",
                    "help": "you can provide axtra options to slurm from this string",
                    "type": "optional",
                },
                SLURM_PREPROCESSING_COMMAND: {
                    "default": "",
                    "help": "python virtual environment to use",
                    "type": "optional",
                },
                SLURM_MODULES_TO_LOADS: {
                    "default": "tomotools/stable",
                    "help": "module to load",
                    "type": "optional",
                },
            },
            NORMALIZATION_BY_SAMPLE_SECTION: {
                NORMALIZATION_BY_SAMPLE_ACTIVE_FIELD: {
                    "default": False,
                    "help": "should we apply frame normalization by a sample or not",
                    "type": "advanced",
                },
                NORMALIZATION_BY_SAMPLE_METHOD: {
                    "default": "median",
                    "help": "method to compute the normalization value",
                    "type": "advanced",
                },
                NORMALIZATION_BY_SAMPLE_SIDE: {
                    "default": "left",
                    "help": "side to pick the sample",
                    "type": "advanced",
                },
                NORMALIZATION_BY_SAMPLE_MARGIN: {
                    "default": 0,
                    "help": "margin (in px) between border and sample",
                    "type": "advanced",
                },
                NORMALIZATION_BY_SAMPLE_WIDTH: {
                    "default": 30,
                    "help": "sample width (in px)",
                    "type": "advanced",
                },
            },
        }

    def to_dict(self):
        """dump configuration to a dict. Must be serializable because might be dump to HDF5 file"""
        return {
            SLURM_SECTION: self.slurm_config.to_dict() if self.slurm_config is not None else SlurmConfig().to_dict(),
            STITCHING_SECTION: {
                STITCHING_TYPE_FIELD: self.stitching_type.value,
                CROSS_CORRELATION_SLICE_FIELD: str(self.slice_for_cross_correlation),
                AXIS_0_POS_PX: _cast_shift_to_str(self.axis_0_pos_px),
                AXIS_0_POS_MM: _cast_shift_to_str(self.axis_0_pos_mm),
                AXIS_0_PARAMS: _dict_to_str(self.axis_0_params or {}),
                AXIS_1_POS_PX: _cast_shift_to_str(self.axis_1_pos_px),
                AXIS_1_POS_MM: _cast_shift_to_str(self.axis_1_pos_mm),
                AXIS_1_PARAMS: _dict_to_str(self.axis_1_params or {}),
                AXIS_2_POS_PX: _cast_shift_to_str(self.axis_2_pos_px),
                AXIS_2_POS_MM: _cast_shift_to_str(self.axis_2_pos_mm),
                AXIS_2_PARAMS: _dict_to_str(self.axis_2_params or {}),
                STITCHING_STRATEGY_FIELD: OverlapStitchingStrategy.from_value(self.stitching_strategy).value,
                FLIP_UD: self.flip_ud,
                FLIP_LR: self.flip_lr,
                RESCALE_FRAMES: self.rescale_frames,
                RESCALE_PARAMS: _dict_to_str(self.rescale_params or {}),
                STITCHING_KERNELS_EXTRA_PARAMS: _dict_to_str(self.stitching_kernels_extra_params or {}),
                AVOID_DATA_DUPLICATION_FIELD: not self.duplicate_data,
            },
            OUTPUT_SECTION: {
                OVERWRITE_RESULTS_FIELD: int(
                    self.overwrite_results,
                ),
            },
            NORMALIZATION_BY_SAMPLE_SECTION: self.normalization_by_sample.to_dict(),
        }


class SingleAxisConfigMetaClass(type):
    """
    Metaclass for single axis stitcher in order to aggregate dumper class and axis

    warning: this class is used by tomwer as well
    """

    def __new__(mcls, name, bases, attrs, axis=None):
        # assert axis is not None
        mcls = super().__new__(mcls, name, bases, attrs)
        mcls._axis = axis
        return mcls


@dataclass
class SingleAxisStitchingConfiguration(StitchingConfiguration, metaclass=SingleAxisConfigMetaClass):
    """
    base class to define z-stitching parameters
    """

    slices: Union[slice, tuple, None] = (
        None  # slices to reconstruct. Over axis 0 for pre-processing, over axis 1 for post-processing. If None will reconstruct all
    )

    alignment_axis_2: AlignmentAxis2 = AlignmentAxis2.CENTER

    pad_mode: str = "constant"  # pad mode to be used for alignment

    @property
    def axis(self) -> int:
        # self._axis is defined by the metaclass
        return self._axis

    def settle_inputs(self) -> None:
        self.settle_slices()

    def settle_slices(self) -> tuple:
        raise ValueError("Base class")

    def get_output_object(self):
        raise ValueError("Base class")

    def to_dict(self):
        if isinstance(self.slices, slice):
            slices = f"{self.slices.start}:{self.slices.stop}:{self.slices.step}"
        elif self.slices in ("", None):
            slices = ""
        else:
            slices = ";".join(str(s) for s in self.slices)
        return concatenate_dict(
            super().to_dict(),
            {
                INPUTS_SECTION: {
                    STITCHING_SLICES: slices,
                },
                STITCHING_SECTION: {
                    ALIGNMENT_AXIS_2_FIELD: self.alignment_axis_2.value,
                    PAD_MODE_FIELD: self.pad_mode,
                },
            },
        )


@dataclass
class PreProcessedSingleAxisStitchingConfiguration(SingleAxisStitchingConfiguration):
    """
    base class to define z-stitching parameters
    """

    input_scans: tuple = ()  # tuple of ScanBase
    output_file_path: str = ""
    output_data_path: str = ""
    output_nexus_version: Optional[float] = None
    pixel_size: Optional[float] = None

    @property
    def stitching_type(self) -> StitchingType:
        if self.axis == 0:
            return StitchingType.Z_PREPROC
        elif self.axis == 1:
            return StitchingType.Y_PREPROC
        else:
            raise ValueError(
                "unexpected axis value. Only stitching over axis 0 (aka z) and 1 (aka y) are handled. Current axis value is %s",
                self.axis,
            )

    def get_output_object(self):
        return NXtomoScan(
            scan=self.output_file_path,
            entry=self.output_data_path,
        )

    def settle_inputs(self) -> None:
        super().settle_inputs()
        self.settle_input_scans()

    def settle_input_scans(self):
        self.input_scans = [
            (
                Factory.create_tomo_object_from_identifier(identifier)
                if isinstance(identifier, (str, ScanIdentifier))
                else identifier
            )
            for identifier in self.input_scans
        ]

    def slice_idx_from_str_to_int(self, index):
        if isinstance(index, str):
            index = index.lower()
            if index == "first":
                return 0
            elif index == "last":
                return len(self.input_scans[0].projections) - 1
            elif index == "middle":
                return max(len(self.input_scans[0].projections) // 2 - 1, 0)
        return int(index)

    def settle_slices(self) -> tuple:
        """
        interpret the slices to be stitched if needed

        Nore: if slices is an instance of slice will redefine start and stop to avoid having negative indexes

        :return: (slices:[slice,Iterable], n_proj:int)
        :rtype: tuple
        """
        slices = self.slices
        if isinstance(slices, Sized) and len(slices) == 0:
            # in this case will stitch them all
            slices = None
        if len(self.input_scans) == 0:
            raise ValueError("No input scan provided")
        if slices is None:
            slices = slice(0, len(self.input_scans[0].projections), 1)
            n_proj = slices.stop
        elif isinstance(slices, slice):
            # force slices indices to be positive
            start = slices.start
            if start < 0:
                start += len(self.input_scans[0].projections) + 1
            stop = slices.stop
            if stop < 0:
                stop += len(self.input_scans[0].projections) + 1
            step = slices.step
            if step is None:
                step = 1
            n_proj = ceil((stop - start) / step)
            # update slices for iteration simplify things
            slices = slice(start, stop, step)
        elif isinstance(slices, (tuple, list)):
            n_proj = len(slices)
            slices = [self.slice_idx_from_str_to_int(s) for s in slices]
        else:
            raise TypeError(f"slices is expected to be a tuple or a lice. Not {type(slices)}")
        self.slices = slices
        return slices, n_proj

    def to_dict(self):
        if self.pixel_size is None:
            pixel_size_mm = ""
        else:
            pixel_size_mm = self.pixel_size * MetricSystem.MILLIMETER.value
        return concatenate_dict(
            super().to_dict(),
            {
                PRE_PROC_SECTION: {
                    DATA_FILE_FIELD: self.output_file_path,
                    DATA_PATH_FIELD: self.output_data_path,
                    NEXUS_VERSION_FIELD: self.output_nexus_version,
                },
                INPUTS_SECTION: {
                    INPUT_DATASETS_FIELD: ";".join(
                        [str(scan.get_identifier()) for scan in self.input_scans],
                    ),
                    INPUT_PIXEL_SIZE_MM: pixel_size_mm,
                },
            },
        )

    @staticmethod
    def get_description_dict() -> dict:
        return concatenate_dict(
            SingleAxisStitchingConfiguration.get_description_dict(),
            {
                PRE_PROC_SECTION: {
                    DATA_FILE_FIELD: {
                        "default": "",
                        "help": "output nxtomo file path",
                        "type": "required",
                    },
                    DATA_PATH_FIELD: {
                        "default": "",
                        "help": "output nxtomo data path",
                        "type": "required",
                    },
                    NEXUS_VERSION_FIELD: {
                        "default": "",
                        "help": "nexus version. If not provided will pick the latest one know",
                        "type": "required",
                    },
                },
            },
        )

    @classmethod
    def from_dict(cls, config: dict):
        if not isinstance(config, dict):
            raise TypeError(f"config is expected to be a dict and not {type(config)}")
        inputs_scans_str = config.get(INPUTS_SECTION, {}).get(INPUT_DATASETS_FIELD, None)
        if inputs_scans_str in (None, ""):
            input_scans = []
        else:
            input_scans = identifiers_as_str_to_instances(inputs_scans_str)

        output_file_path = config.get(PRE_PROC_SECTION, {}).get(DATA_FILE_FIELD, None)

        nexus_version = config.get(PRE_PROC_SECTION, {}).get(NEXUS_VERSION_FIELD, None)
        if nexus_version in (None, ""):
            nexus_version = nxtomo.LATEST_VERSION
        else:
            nexus_version = float(nexus_version)
        pixel_size = config.get(INPUT_PIXEL_SIZE_MM, "").replace(" ", "")
        if pixel_size == "":
            pixel_size = None
        else:
            pixel_size = float(pixel_size) / MetricSystem.MM

        return cls(
            stitching_strategy=OverlapStitchingStrategy.from_value(
                config[STITCHING_SECTION].get(
                    STITCHING_STRATEGY_FIELD,
                    OverlapStitchingStrategy.COSINUS_WEIGHTS,
                ),
            ),
            axis_0_pos_px=str_to_shifts(config[STITCHING_SECTION].get(AXIS_0_POS_PX, None)),
            axis_0_pos_mm=str_to_shifts(config[STITCHING_SECTION].get(AXIS_0_POS_MM, None)),
            axis_0_params=_valid_shifts_params(_str_to_dict(config[STITCHING_SECTION].get(AXIS_0_PARAMS, {}))),
            axis_1_pos_px=str_to_shifts(config[STITCHING_SECTION].get(AXIS_1_POS_PX, None)),
            axis_1_pos_mm=str_to_shifts(config[STITCHING_SECTION].get(AXIS_1_POS_MM, None)),
            axis_1_params=_valid_shifts_params(
                _str_to_dict(
                    config[STITCHING_SECTION].get(AXIS_1_PARAMS, {}),
                )
            ),
            axis_2_pos_px=str_to_shifts(config[STITCHING_SECTION].get(AXIS_2_POS_PX, None)),
            axis_2_pos_mm=str_to_shifts(config[STITCHING_SECTION].get(AXIS_2_POS_MM, None)),
            axis_2_params=_valid_shifts_params(
                _str_to_dict(
                    config[STITCHING_SECTION].get(AXIS_2_PARAMS, {}),
                )
            ),
            input_scans=input_scans,
            output_file_path=output_file_path,
            output_data_path=config.get(PRE_PROC_SECTION, {}).get(DATA_PATH_FIELD, "entry_from_stitchig"),
            overwrite_results=config[STITCHING_SECTION].get(OVERWRITE_RESULTS_FIELD, True),
            output_nexus_version=nexus_version,
            slices=_slices_to_list_or_slice(config[INPUTS_SECTION].get(STITCHING_SLICES, None)),
            slurm_config=SlurmConfig.from_dict(config.get(SLURM_SECTION, {})),
            slice_for_cross_correlation=config[STITCHING_SECTION].get(CROSS_CORRELATION_SLICE_FIELD, "middle"),
            pixel_size=pixel_size,
            flip_ud=_scalar_or_tuple_to_bool_or_tuple_of_bool(config[STITCHING_SECTION].get(FLIP_UD, False)),
            flip_lr=_scalar_or_tuple_to_bool_or_tuple_of_bool(config[STITCHING_SECTION].get(FLIP_LR, False)),
            rescale_frames=convert_to_bool(config[STITCHING_SECTION].get(RESCALE_FRAMES, 0))[0],
            rescale_params=_str_to_dict(config[STITCHING_SECTION].get(RESCALE_PARAMS, {})),
            stitching_kernels_extra_params=_valid_stitching_kernels_params(
                _str_to_dict(
                    config[STITCHING_SECTION].get(STITCHING_KERNELS_EXTRA_PARAMS, {}),
                )
            ),
            alignment_axis_2=AlignmentAxis2.from_value(
                config[STITCHING_SECTION].get(ALIGNMENT_AXIS_2_FIELD, AlignmentAxis2.CENTER)
            ),
            pad_mode=config[STITCHING_SECTION].get(PAD_MODE_FIELD, "constant"),
            duplicate_data=not config[STITCHING_SECTION].get(AVOID_DATA_DUPLICATION_FIELD, False),
            normalization_by_sample=NormalizationBySample.from_dict(config.get(NORMALIZATION_BY_SAMPLE_SECTION, {})),
        )


@dataclass
class PostProcessedSingleAxisStitchingConfiguration(SingleAxisStitchingConfiguration):
    """
    base class to define z-stitching parameters
    """

    input_volumes: tuple = ()  # tuple of VolumeBase
    output_volume: Optional[VolumeIdentifier] = None
    voxel_size: Optional[float] = None
    alignment_axis_1: AlignmentAxis1 = AlignmentAxis1.CENTER

    @property
    def stitching_type(self) -> StitchingType:
        if self.axis == 0:
            return StitchingType.Z_POSTPROC
        else:
            raise ValueError(f"unexpected axis value. Only stitching over axis 0 (aka z) is handled. Not {self.axis}")

    def get_output_object(self):
        return self.output_volume

    def settle_inputs(self) -> None:
        super().settle_inputs()
        self.settle_input_volumes()

    def settle_input_volumes(self):
        self.input_volumes = [
            (
                Factory.create_tomo_object_from_identifier(identifier)
                if isinstance(identifier, (str, VolumeIdentifier))
                else identifier
            )
            for identifier in self.input_volumes
        ]

    def slice_idx_from_str_to_int(self, index):
        if isinstance(index, str):
            index = index.lower()
            if index == "first":
                return 0
            elif index == "last":
                return self.input_volumes[0].get_volume_shape()[1] - 1
            elif index == "middle":
                return max(self.input_volumes[0].get_volume_shape()[1] // 2 - 1, 0)
        return int(index)

    def settle_slices(self) -> tuple:
        """
        interpret the slices to be stitched if needed

        Nore: if slices is an instance of slice will redefine start and stop to avoid having negative indexes

        :return: (slices:[slice,Iterable], n_proj:int)
        :rtype: tuple
        """
        slices = self.slices
        if isinstance(slices, Sized) and len(slices) == 0:
            # in this case will stitch them all
            slices = None
        if len(self.input_volumes) == 0:
            raise ValueError("No input volume provided. Cannot settle slices")
        if slices is None:
            # before alignment was existing
            # slices = slice(0, self.input_volumes[0].get_volume_shape()[1], 1)
            slices = slice(
                0,
                max([volume.get_volume_shape()[1] for volume in self.input_volumes]),
                1,
            )
            n_slices = slices.stop
        if isinstance(slices, slice):
            # force slices indices to be positive
            start = slices.start
            if start < 0:
                start += max([volume.get_volume_shape()[1] for volume in self.input_volumes]) + 1
            stop = slices.stop
            if stop < 0:
                stop += max([volume.get_volume_shape()[1] for volume in self.input_volumes]) + 1
            step = slices.step
            if step is None:
                step = 1
            n_slices = ceil((stop - start) / step)
            # update slices for iteration simplify things
            slices = slice(start, stop, step)
        elif isinstance(slices, Iterable):
            n_slices = len(slices)
            slices = [self.slice_idx_from_str_to_int(s) for s in slices]
        else:
            raise TypeError(f"slices is expected to be a tuple or a slice. Not {type(slices)}")
        self.slices = slices
        return slices, n_slices

    @classmethod
    def from_dict(cls, config: dict):
        if not isinstance(config, dict):
            raise TypeError(f"config is expected to be a dict and not {type(config)}")
        inputs_volumes_str = config.get(INPUTS_SECTION, {}).get(INPUT_DATASETS_FIELD, None)
        if inputs_volumes_str in (None, ""):
            input_volumes = []
        else:
            input_volumes = identifiers_as_str_to_instances(inputs_volumes_str)
        overwrite_results = config[STITCHING_SECTION].get(OVERWRITE_RESULTS_FIELD, True) in ("1", True, "True", 1)
        output_volume = config.get(POST_PROC_SECTION, {}).get(OUTPUT_VOLUME, None)
        if output_volume is not None:
            output_volume = Factory.create_tomo_object_from_identifier(output_volume)
            output_volume.overwrite = overwrite_results
        voxel_size = config.get(INPUTS_SECTION, {}).get(INPUT_VOXEL_SIZE_MM, "")
        voxel_size = voxel_size.replace(" ", "")
        if voxel_size == "":
            voxel_size = None
        else:
            voxel_size = float(voxel_size) * MetricSystem.MM

        # on the next section the one with a default value qre the optional one
        return cls(
            stitching_strategy=OverlapStitchingStrategy.from_value(
                config[STITCHING_SECTION].get(
                    STITCHING_STRATEGY_FIELD,
                    OverlapStitchingStrategy.COSINUS_WEIGHTS,
                ),
            ),
            axis_0_pos_px=str_to_shifts(config[STITCHING_SECTION].get(AXIS_0_POS_PX, None)),
            axis_0_pos_mm=str_to_shifts(config[STITCHING_SECTION].get(AXIS_0_POS_MM, None)),
            axis_0_params=_valid_shifts_params(config[STITCHING_SECTION].get(AXIS_0_PARAMS, {})),
            axis_1_pos_px=str_to_shifts(config[STITCHING_SECTION].get(AXIS_1_POS_PX, None)),
            axis_1_pos_mm=str_to_shifts(config[STITCHING_SECTION].get(AXIS_1_POS_MM, None)),
            axis_1_params=_valid_shifts_params(config[STITCHING_SECTION].get(AXIS_1_PARAMS, {})),
            axis_2_pos_px=str_to_shifts(config[STITCHING_SECTION].get(AXIS_2_POS_PX, None)),
            axis_2_pos_mm=str_to_shifts(config[STITCHING_SECTION].get(AXIS_2_POS_MM, None)),
            axis_2_params=_valid_shifts_params(config[STITCHING_SECTION].get(AXIS_2_PARAMS, {})),
            input_volumes=input_volumes,
            output_volume=output_volume,
            overwrite_results=overwrite_results,
            slices=_slices_to_list_or_slice(config[INPUTS_SECTION].get(STITCHING_SLICES, None)),
            slurm_config=SlurmConfig.from_dict(config.get(SLURM_SECTION, {})),
            voxel_size=voxel_size,
            slice_for_cross_correlation=config[STITCHING_SECTION].get(CROSS_CORRELATION_SLICE_FIELD, "middle"),
            flip_ud=_scalar_or_tuple_to_bool_or_tuple_of_bool(config[STITCHING_SECTION].get(FLIP_UD, False)),
            flip_lr=_scalar_or_tuple_to_bool_or_tuple_of_bool(config[STITCHING_SECTION].get(FLIP_LR, False)),
            rescale_frames=convert_to_bool(config[STITCHING_SECTION].get(RESCALE_FRAMES, 0))[0],
            rescale_params=_str_to_dict(config[STITCHING_SECTION].get(RESCALE_PARAMS, {})),
            stitching_kernels_extra_params=_valid_stitching_kernels_params(
                _str_to_dict(
                    config[STITCHING_SECTION].get(STITCHING_KERNELS_EXTRA_PARAMS, {}),
                )
            ),
            alignment_axis_1=AlignmentAxis1.from_value(
                config[STITCHING_SECTION].get(ALIGNMENT_AXIS_1_FIELD, AlignmentAxis1.CENTER)
            ),
            alignment_axis_2=AlignmentAxis2.from_value(
                config[STITCHING_SECTION].get(ALIGNMENT_AXIS_2_FIELD, AlignmentAxis2.CENTER)
            ),
            pad_mode=config[STITCHING_SECTION].get(PAD_MODE_FIELD, "constant"),
            duplicate_data=not config[STITCHING_SECTION].get(AVOID_DATA_DUPLICATION_FIELD, False),
            normalization_by_sample=NormalizationBySample.from_dict(config.get(NORMALIZATION_BY_SAMPLE_SECTION, {})),
        )

    def to_dict(self):
        if self.voxel_size is None:
            voxel_size_mm = ""
        else:
            voxel_size_mm = numpy.array(self.voxel_size) / MetricSystem.MM

        return concatenate_dict(
            super().to_dict(),
            {
                INPUTS_SECTION: {
                    INPUT_DATASETS_FIELD: [volume.get_identifier().to_str() for volume in self.input_volumes],
                    INPUT_VOXEL_SIZE_MM: voxel_size_mm,
                },
                POST_PROC_SECTION: {
                    OUTPUT_VOLUME: (
                        self.output_volume.get_identifier().to_str() if self.output_volume is not None else ""
                    ),
                },
                STITCHING_SECTION: {
                    ALIGNMENT_AXIS_1_FIELD: self.alignment_axis_1.value,
                },
            },
        )

    @staticmethod
    def get_description_dict() -> dict:
        return concatenate_dict(
            SingleAxisStitchingConfiguration.get_description_dict(),
            {
                POST_PROC_SECTION: {
                    OUTPUT_VOLUME: {
                        "default": "",
                        "help": "identifier of the output volume. Like hdf5:volume:[file_path]?path=[data_path] for an HDF5 volume",
                        "type": "required",
                    },
                },
                STITCHING_SECTION: {
                    ALIGNMENT_AXIS_1_FIELD: {
                        "default": "center",
                        "help": f"alignment to apply over axis 1 if needed. Valid values are {AlignmentAxis1.values()}",
                        "type": "advanced",
                    }
                },
            },
        )


def identifiers_as_str_to_instances(list_identifiers_as_str: str) -> tuple:
    # convert str to a list of str that should represent identifiers
    if isinstance(list_identifiers_as_str, str):
        list_identifiers_as_str = list_identifiers_as_str.lstrip("[").lstrip("(")
        list_identifiers_as_str = list_identifiers_as_str.rstrip("]").rstrip(")")
        identifiers_as_str = convert_str_to_tuple(list_identifiers_as_str.replace(";", ","))
    else:
        identifiers_as_str = list_identifiers_as_str
    if identifiers_as_str is None:
        return tuple()
    # convert identifiers as string to IdentifierType instances
    return tuple(
        [Factory.create_tomo_object_from_identifier(identifier_as_str) for identifier_as_str in identifiers_as_str]
    )


def dict_to_config_obj(config: dict):
    if not isinstance(config, dict):
        raise TypeError
    stitching_type = config.get(STITCHING_SECTION, {}).get(STITCHING_TYPE_FIELD, None)
    if stitching_type is None:
        raise ValueError("Unable to find stitching type from config dict")
    else:
        stitching_type = StitchingType.from_value(stitching_type)
        if stitching_type is StitchingType.Z_POSTPROC:
            return PostProcessedZStitchingConfiguration.from_dict(config)
        elif stitching_type is StitchingType.Z_PREPROC:
            return PreProcessedZStitchingConfiguration.from_dict(config)
        elif stitching_type is StitchingType.Y_PREPROC:
            return PreProcessedYStitchingConfiguration.from_dict(config)
        else:
            raise NotImplementedError(f"stitching type {stitching_type.value} not handled yet")


def get_default_stitching_config(stitching_type: Optional[Union[StitchingType, str]]) -> tuple:
    """
    Return a default configuration for doing stitching.

    :param stitching_type: if None then return a configuration were use can provide inputs for any
                           of the stitching.
                           Else return config dict dedicated to a particular stitching
    :return: (config, section comments)
    """
    if stitching_type is None:
        return concatenate_dict(z_postproc_stitching_config, z_preproc_stitching_config)

    stitching_type = StitchingType.from_value(stitching_type)
    if stitching_type is StitchingType.Z_POSTPROC:
        return z_postproc_stitching_config
    elif stitching_type is StitchingType.Z_PREPROC:
        return z_preproc_stitching_config
    elif stitching_type is StitchingType.Y_PREPROC:
        return y_preproc_stitching_config
    else:
        raise NotImplementedError


class PreProcessedYStitchingConfiguration(PreProcessedSingleAxisStitchingConfiguration, axis=1):
    pass


class PreProcessedZStitchingConfiguration(PreProcessedSingleAxisStitchingConfiguration, axis=0):
    pass


class PostProcessedZStitchingConfiguration(PostProcessedSingleAxisStitchingConfiguration, axis=0):
    pass


y_preproc_stitching_config = PreProcessedYStitchingConfiguration.get_description_dict()

z_preproc_stitching_config = PreProcessedZStitchingConfiguration.get_description_dict()

z_postproc_stitching_config = PostProcessedZStitchingConfiguration.get_description_dict()



import numpy as np
from scipy.interpolate import interp1d
from ..utils import get_2D_3D_shape, check_supported, deprecated_class, deprecated


class SinoBuilder:
    """
    A class to build sinograms.
    """

    def __init__(
        self, sinos_shape=None, radios_shape=None, rot_center=None, halftomo=False, angles=None, interpolate=False
    ):
        """
        Initialize a SinoBuilder instance.

        Parameters
        ----------
        sinos_shape: tuple of int
            Shape of the stack of sinograms, in the form `(n_z, n_angles, n_x)`.
            If not provided, it is derived from `radios_shape`.
        radios_shape: tuple of int
            Shape of the chunk of radios, in the form `(n_angles, n_z, n_x)`.
            If not provided, it is derived from `sinos_shape`.
        rot_center: int or array
            Rotation axis position. A scalar indicates the same rotation axis position
            for all the projections.
        halftomo: bool
            Whether "half tomography" is enabled. Default is False.
        interpolate: bool, optional
            Only used if halftomo=True.
            Whether to re-grid the second part of sinograms to match projection k with projection k + n_a/2.
            This forces each pair of projection (k, k + n_a/2) to be separated by exactly 180 degrees.
        angles: array, optional
            Rotation angles (in radians). Used and required only when halftomo and interpolate are True.
        """
        self._get_shapes(sinos_shape, radios_shape)
        self.set_rot_center(rot_center)
        self._configure_halftomo(halftomo, interpolate, angles)

    def _get_shapes(self, sinos_shape, radios_shape=None):
        if (sinos_shape is None) and (radios_shape is None):
            raise ValueError("Need to provide sinos_shape and/or radios_shape")
        if sinos_shape is None:
            n_a, n_z, n_x = get_2D_3D_shape(radios_shape)
            sinos_shape = (n_z, n_a, n_x)
        elif len(sinos_shape) == 2:
            sinos_shape = (1,) + sinos_shape
        if radios_shape is None:
            n_z, n_a, n_x = get_2D_3D_shape(sinos_shape)
            radios_shape = (n_a, n_z, n_x)
        elif len(radios_shape) == 2:
            radios_shape = (1,) + radios_shape

        self.sinos_shape = sinos_shape
        self.radios_shape = radios_shape
        n_a, n_z, n_x = radios_shape
        self.n_angles = n_a
        self.n_z = n_z
        self.n_x = n_x

    def set_rot_center(self, rot_center):
        """
        Set the rotation axis position for the current radios/sinos stack.

        rot_center: int or array
            Rotation axis position. A scalar indicates the same rotation axis position
            for all the projections.
        """
        if rot_center is None:
            rot_center = (self.n_x - 1) / 2.0
        if not (np.isscalar(rot_center)):
            rot_center = np.array(rot_center)
            if rot_center.size != self.n_angles:
                raise ValueError(
                    "Expected rot_center to have %d elements but got %d" % (self.n_angles, rot_center.size)
                )
        self.rot_center = rot_center

    def _configure_halftomo(self, halftomo, interpolate, angles):
        self.halftomo = halftomo
        self.interpolate = interpolate
        self.angles = angles
        self._halftomo_flip = False
        if not self.halftomo:
            return
        if interpolate and (angles is None):
            raise ValueError("The parameter 'angles' has to be provided when using halftomo=True and interpolate=True")
        self.extended_sino_width = get_extended_sinogram_width(self.n_x, self.rot_center)
        # If CoR is on the left: "flip" the logic
        if self.rot_center < (self.n_x - 1) / 2:
            self.rot_center = self.n_x - 1 - self.rot_center
            self._halftomo_flip = True
        #

        if abs(self.rot_center - ((self.n_x - 1) / 2.0)) < 1:  # which tol ?
            raise ValueError("Half tomography: incompatible rotation axis position: %.2f" % self.rot_center)
        self.sinos_halftomo_shape = (self.n_z, (self.n_angles + 1) // 2, self.extended_sino_width)

    def _check_array_shape(self, array, kind="radio"):
        expected_shape = self.radios_shape if "radio" in kind else self.sinos_shape
        assert array.shape == expected_shape, "Expected radios shape %s, but got %s" % (expected_shape, array.shape)

    @property
    def output_shape(self):
        """
        Get the output sinograms shape.
        """
        if self.halftomo:
            return self.sinos_halftomo_shape
        return self.sinos_shape

    #
    # 2D
    #

    def _get_sino_simple(self, radios, i):
        return radios[:, i, :]  # view

    def _get_sino_halftomo(self, sino, output=None):
        # TODO output is ignored for now
        if self.interpolate:
            match_half_sinos_parts(sino, self.angles)
        elif self.n_angles & 1:
            # Odd number of projections - add one line in the end
            sino = np.vstack([sino, np.zeros_like(sino[-1])])
        if self._halftomo_flip:
            sino = sino[:, ::-1]

        if self.rot_center > self.n_x:
            # (hopefully rare) case where CoR is outside FoV
            result = _convert_halftomo_right(sino, self.extended_sino_width)
        else:
            # Standard case
            result = convert_halftomo(sino, self.extended_sino_width)

        if self._halftomo_flip:
            result = result[:, ::-1]
        return result

    def get_sino(self, radios, i, output=None):
        """
        The the sinogram at a given index.

        Parameters
        ----------
        radios: array
            3D array with shape (n_z, n_angles, n_x)
        i: int
            Sinogram index

        Returns
        -------
        sino: array
            Two dimensional array with shape (n_angles2, n_x2) where the dimensions
            are determined by the current settings.
        """
        sino = self._get_sino_simple(radios, i)
        if self.halftomo:
            return self._get_sino_halftomo(sino, output=None)
        else:
            return sino

    def convert_sino(self, sino, output=None):
        if not self.halftomo:
            return sino
        return self._get_sino_halftomo(sino, output=output)

    #
    # 3D
    #

    def _get_sinos_simple(self, radios, output=None):
        res = np.rollaxis(radios, 1, 0)  # view
        if output is not None:
            output[...] = res[...]  # copy
            return output
        return res

    def _get_sinos_halftomo(self, radios, output=None):
        n_a, n_z, n_x = radios.shape
        if output is None:
            output = np.zeros(self.sinos_halftomo_shape, dtype=np.float32)
        elif output.shape != self.output_shape:
            raise ValueError("Expected output shape to be %s, but got %s" % (str(output.shape), str(self.output_shape)))
        for i in range(n_z):
            sino = self._get_sino_simple(radios, i)
            output[i] = self._get_sino_halftomo(sino)
        return output

    def get_sinos(self, radios, output=None):
        if self.halftomo:
            return self._get_sinos_halftomo(radios, output=output)
        else:
            return self._get_sinos_simple(radios, output=output)

    @deprecated("Use get_sino() or get_sinos() instead", do_print=True)
    def radios_to_sinos(self, radios, output=None, copy=False):
        """
        DEPRECATED. Use get_sinos() or get_sino() instead.
        """
        return self.get_sinos(radios, output=output)


SinoProcessing = deprecated_class("'SinoProcessing' was renamed 'SinoBuilder'", do_print=True)(SinoBuilder)


class SinoMult:
    """
    A class for preparing a sinogram for half-tomography reconstruction, without stitching the two parts
    """

    def __init__(self, sino_shape, rot_center):
        self._set_shape(sino_shape)
        self._prepare_weights(rot_center)

    def _set_shape(self, sino_shape):
        _, self.n_a, self.n_x = get_2D_3D_shape(sino_shape)

    def _prepare_weights(self, rot_center):
        n_x = self.n_x
        middle = (n_x - 1) / 2.0
        if rot_center >= middle:
            overlap_width = int(2 * (n_x - 1 - rot_center))
            self.overlap_region = slice(-overlap_width, None)
            self.pad_left, self.pad_right = 0, n_x - overlap_width
        else:
            overlap_width = int(2 * rot_center)
            self.overlap_region = slice(0, overlap_width)
            self.pad_left, self.pad_right = n_x - overlap_width, 0
        weights = np.linspace(0, 1, overlap_width, endpoint=True)
        if rot_center >= middle:
            weights = weights[::-1]
        self.weights = np.ascontiguousarray(weights, dtype="f")
        overlap_region_indices = np.arange(self.n_x)[self.overlap_region]
        self.start_x = overlap_region_indices[0]
        self.end_x = overlap_region_indices[-1]
        self.extended_width = n_x + self.pad_left + self.pad_right

    def prepare_sino(self, sino):
        sino[:, self.overlap_region] *= self.weights
        return sino


def convert_halftomo(sino, extended_width, transition_width=None):
    """
    Converts a sinogram into a sinogram with extended FOV with the "half tomography"
    setting.
    """
    assert sino.ndim == 2
    assert (sino.shape[0] % 2) == 0
    na, nx = sino.shape

    na2 = na // 2
    r = extended_width // 2
    d = transition_width or nx - r
    res = np.zeros((na2, 2 * r), dtype="f")

    sino1 = sino[:na2, :]
    sino2 = sino[na2:, ::-1]
    res[:, : r - d] = sino1[:, : r - d]
    #
    w1 = np.linspace(0, 1, 2 * d, endpoint=True)
    res[:, r - d : r + d] = (1 - w1) * sino1[:, r - d :] + w1 * sino2[:, 0 : 2 * d]
    #
    res[:, r + d :] = sino2[:, 2 * d :]

    return res


# This function can have a cuda counterpart, see test_interpolation.py
def match_half_sinos_parts(sino, angles, output=None):
    """
    Modifies the lower part of the half-acquisition sinogram so that each projection pair is
    separated by exactly 180 degrees.
    This means that `new_sino[k]` and `new_sino[k + n_angles//2]` will be 180 degrees apart.

    Parameters
    ----------
    sino: numpy.ndarray
        Two dimensional array with the sinogram in the form (n_angles, n_x)
    angles: numpy.ndarray
        One dimensional array with the rotation angles.
    output: numpy.array, optional
        Output sinogram. By default, the array 'sino' is modified in-place.

    Notes
    -----
    This function assumes that the angles are in an increasing order.
    """
    n_a = angles.size
    n_a_2 = n_a // 2
    sino_part1 = sino[:n_a_2, :]
    sino_part2 = sino[n_a_2:, :]
    angles = np.rad2deg(angles)  # more numerically stable ?
    angles_1 = angles[:n_a_2]
    angles_2 = angles[n_a_2:]
    angles_2_target = angles_1 + 180.0
    interpolator = interp1d(angles_2, sino_part2, axis=0, kind="linear", copy=False, fill_value="extrapolate")
    if output is None:
        output = sino
    else:
        output[:n_a_2, :] = sino[:n_a_2, :]
    output[n_a_2:, :] = interpolator(angles_2_target)
    return output


# EXPERIMENTAL
def _convert_halftomo_right(sino, extended_width):
    """
    Converts a sinogram into a sinogram with extended FOV with the "half tomography"
    setting, with a CoR outside the image support.
    """
    assert sino.ndim == 2
    na, nx = sino.shape
    assert (na % 2) == 0
    rotation_axis_position = extended_width // 2
    assert rotation_axis_position > nx

    sino2 = np.pad(sino, ((0, 0), (0, rotation_axis_position - nx)), mode="reflect")
    return convert_halftomo(sino2, extended_width)


def get_extended_sinogram_width(sino_width, rotation_axis_position):
    """
    Compute the width (in pixels) of the extended sinogram for half-acquisition setting.
    """
    middle = (sino_width - 1) / 2.0
    if rotation_axis_position >= middle:
        overlap_width = int(2 * (sino_width - 1 - rotation_axis_position))
    else:
        overlap_width = int(2 * rotation_axis_position)
    return 2 * sino_width - overlap_width


def prepare_half_tomo_sinogram(sino, rot_center, get_extended_sino=True):
    if get_extended_sino:
        sino = sino.copy()
    n_angles, n_x = sino.shape
    middle = (n_x - 1) / 2.0
    if rot_center >= middle:
        overlap_width = int(2 * (n_x - 1 - rot_center))
        overlap_region = slice(-overlap_width, None)
        pad_left, pad_right = 0, n_x - overlap_width
    else:
        overlap_width = int(2 * rot_center)
        overlap_region = slice(0, overlap_width)
        pad_left, pad_right = n_x - overlap_width, 0
    weights = np.linspace(0, 1, overlap_width, endpoint=True)
    if rot_center >= middle:
        weights = weights[::-1]
    sino[:, overlap_region] *= weights
    if get_extended_sino:
        return np.pad(sino, ((0, 0), (pad_left, pad_right)), mode="constant")
    return sino


class SinoNormalization:
    """
    A class for sinogram normalization utilities.
    """

    kinds = [
        "chebyshev",
        "subtraction",
        "division",
    ]
    operations = {"subtraction": np.subtract, "division": np.divide}

    def __init__(self, kind="chebyshev", sinos_shape=None, radios_shape=None, normalization_array=None):
        """
        Initialize a SinoNormalization class.

        Parameters
        -----------
        kind: str, optional
            Normalization type. They can be the following:
               - chebyshev: Each sinogram line is estimated by a Chebyshev polynomial
               of degree 2. This estimation is then subtracted from the sinogram.
               - subtraction: Each sinogram is subtracted with a user-provided array.
                 The array can be 1D (angle-independent) and 2D (angle-dependent)
               - division: same as previously, but with a division operation.
            Default is "chebyshev"
        sinos_shape: tuple, optional
            Shape of the sinogram or sinogram stack.
            Either this parameter or 'radios_shape' has to be provided.
        radios_shape: tuple, optional
            Shape of the projections or projections stack.
            Either this parameter or 'sinos_shape' has to be provided.
        normalization_array: numpy.ndarray, optional
            Normalization array when kind='subtraction' or kind='division'.
        """
        self._get_shapes(sinos_shape, radios_shape)
        self._set_kind(kind, normalization_array)

    _get_shapes = SinoBuilder._get_shapes

    def _set_kind(self, kind, normalization_array):
        check_supported(kind, self.kinds, "sinogram normalization kind")
        self.normalization_kind = kind
        self._normalization_instance_method = self._normalize_chebyshev  # default
        if kind in ["subtraction", "division"]:
            if not isinstance(normalization_array, np.ndarray):
                raise ValueError(
                    "Expected 'normalization_array' to be provided as a numpy array for normalization kind='%s'" % kind
                )
            if normalization_array.shape[-1] != self.sinos_shape[-1]:
                n_a, n_x = self.sinos_shape[-2:]
                raise ValueError("Expected normalization_array to have shape (%d, %d) or (%d, )" % (n_a, n_x, n_x))
            self.norm_operation = self.operations[kind]
            self._normalization_instance_method = self._normalize_op
        self.normalization_array = normalization_array

    #
    # Chebyshev normalization
    #

    def _normalize_chebyshev_2D(self, sino):
        output = sino  # inplace
        Nr, Nc = sino.shape
        J = np.arange(Nc)
        x = 2.0 * (J + 0.5 - Nc / 2) / Nc
        sum0 = Nc
        f2 = 3.0 * x * x - 1.0
        sum1 = (x**2).sum()
        sum2 = (f2**2).sum()
        for i in range(Nr):
            ff0 = sino[i, :].sum()
            ff1 = (x * sino[i, :]).sum()
            ff2 = (f2 * sino[i, :]).sum()
            output[i, :] = sino[i, :] - (ff0 / sum0 + ff1 * x / sum1 + ff2 * f2 / sum2)
        return output

    def _normalize_chebyshev_3D(self, sino):
        for i in range(sino.shape[0]):
            self._normalize_chebyshev_2D(sino[i])
        return sino

    def _normalize_chebyshev(self, sino):
        if sino.ndim == 2:
            self._normalize_chebyshev_2D(sino)
        else:
            self._normalize_chebyshev_3D(sino)
        return sino

    #
    # Array subtraction/division
    #

    def _normalize_op(self, sino):
        if sino.ndim == 2:
            self.norm_operation(sino, self.normalization_array, out=sino)
        else:
            for i in range(sino.shape[0]):
                self.norm_operation(sino[i], self.normalization_array, out=sino[i])
        return sino

    #
    # Dispatch
    #

    def normalize(self, sino):
        """
        Normalize a sinogram or stack of sinogram.
        The process is done in-place, meaning that the sinogram content is overwritten.
        """
        return self._normalization_instance_method(sino)



from ..processing.histogram import *
from ..utils import deprecation_warning

deprecation_warning(
    "nabu.misc.histogram has been moved to nabu.processing.histogram", do_print=True, func_name="histogram"
)
